#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test033_rotation_spect_aa_helpers import *

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()
sources = create_test(sim)

# AA mode
for source in sources:
    source.direction.acceptance_angle.intersection_flag = False
    source.direction.acceptance_angle.normal_flag = False

# go
output = sim.start()

# check
is_ok = evaluate_test(output, sources, 12, 0)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test038_gan_phsp_spect_gan_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test038_gan_phsp_spect")
paths.output_ref = paths.output_ref / "test038_ref"

# create the simulation
sim = gate.Simulation()
condition_generator = create_simulation(sim, paths)

gsource = sim.get_source_user_info("gaga")
gsource.skip_policy = "SkipEvents"  # this is SkipEvents by default

# go (cannot be spawn in another process)
output = sim.start(False)

# test
all_cond = condition_generator.all_cond
analyze_results(output, paths, all_cond)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test028_ge_nm670_spect_2_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test028_ge_nm670_spect")

# create the simulation
sim = gate.Simulation()

# main description
spect = create_spect_simu(sim, paths, 1)
test_add_proj(sim, paths)

# rotate spect
cm = gate.g4_units("cm")
psd = 6.11 * cm
p = [0, 0, -(20 * cm + psd)]
spect.translation, spect.rotation = gate.get_transform_orbiting(p, "y", -15)

sec = gate.g4_units("second")
sim.run_timing_intervals = [[1 * sec, 2 * sec]]

output = sim.start()

# check
is_ok = test_spect_hits(output, paths, version="3")

# check
proj = output.get_actor("Projection")
is_ok = test_spect_proj(output, paths, proj, version="3") and is_ok

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test028_ge_nm670_spect_4_acc_angle_helpers import *

# create the simulation
sim = gate.Simulation()

# simu description
spect, proj = create_spect_simu(
    sim,
    paths,
    number_of_threads=1,
    activity_kBq=1000,
    aa_enabled=True,
    aa_mode="ZeroEnergy",
)

# go
print(proj)
output = sim.start(True)
print(proj)

# check
is_ok = compare_result(output, proj, "test028_aa_zero_energy.png", sum_tolerance=17)
gate.test_ok(is_ok)



import opengate as gate


class SolidBuilderBase:
    def init_user_info(self, user_info):
        pass

    def Build(self, user_info):
        s = f"You must implement the Build function, and return a G4Solid"
        gate.fatal(s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import os

paths = gate.get_default_test_paths(__file__, "gate_test044_pbs")

particle = "Carbon_"
energy = "1440MeV_"
beam_shape = "sourceShapePBS"
folder = particle + energy + beam_shape

output_path = paths.output / "output_test044"
ref_path = paths.gate_output

# for the for loop
start = -500
spacing = 100
end = -start + spacing
planePositionsV = range(start, end, spacing)

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_seed = 123654789
ui.random_engine = "MersenneTwister"

# units
km = gate.g4_units("km")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
um = gate.g4_units("um")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
nm = gate.g4_units("nm")
deg = gate.g4_units("deg")
mrad = gate.g4_units("mrad")

# add a material database
sim.add_material_database(paths.gate_data / "HFMaterials2014.db")

#  change world size
world = sim.world
world.size = [600 * cm, 500 * cm, 500 * cm]

# waterbox
# translation and rotation like in the Gate macro
waterbox = sim.add_volume("Box", "waterbox")
waterbox.size = [10 * cm, 10 * cm, 100.2 * cm]
waterbox.translation = [0 * cm, 0 * cm, 50.1 * cm]
waterbox.material = "Vacuum"
waterbox.color = [0, 0, 1, 1]

# Planes
m = Rotation.identity().as_matrix()

for i in planePositionsV:
    plane = sim.add_volume("Box", "planeNr" + str(i) + "a")
    plane.mother = "waterbox"
    plane.size = [100 * mm, 100 * mm, 2 * mm]
    plane.translation = [0 * mm, 0 * mm, i * mm]
    plane.rotation = m
    plane.material = "G4_AIR"
    plane.color = [1, 0, 1, 1]

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "FTFP_INCLXX_EMZ"
sim.set_cut("world", "all", 1000 * km)

# default source for tests (from test42)
source = sim.add_source("PencilBeamSource", "mysource")
source.energy.mono = 1440 * MeV
source.particle = "ion 6 12"  # carbon
source.position.type = "disc"  # pos = Beam, shape = circle + sigma
source.position.translation = [0 * mm, 0 * mm, 0 * cm]
# rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.n = 20000
source.direction.partPhSp_x = [
    2.3335754 * mm,
    2.3335754 * mrad,
    0.00078728 * mm * mrad,
    0,
]
source.direction.partPhSp_y = [
    1.96433431 * mm,
    0.00079118 * mrad,
    0.00249161 * mm * mrad,
    0,
]

count = 0
# add dose actors
for i in planePositionsV:
    dose = sim.add_actor("DoseActor", "doseInYZ" + str(i))
    filename = "plane" + str(i) + "a.mhd"
    dose.output = output_path / filename
    dose.mother = "planeNr" + str(i) + "a"
    dose.size = [250, 250, 1]
    dose.spacing = [0.4, 0.4, 2]
    dose.hit_type = "random"
    count += 1

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# create output dir, if it doesn't exist
if not os.path.isdir(output_path):
    os.mkdir(output_path)

# start simulation
output = sim.start()

# print results at the end
stat = output.get_actor("Stats")
print(stat)

print("Start to analyze data")
override = False
if (not os.path.exists(ref_path / "sigma_values.txt")) or override:
    sigmasRef, musRef = gate.write_gauss_param_to_file(
        ref_path,
        planePositionsV,
        saveFig=False,
        fNamePrefix="plane",
        fNameSuffix="a_Carbon_1440MeV_sourceShapePBS-Edep.mhd",
    )
override = False
if (not os.path.exists(output_path / "sigma_values.txt")) or override:
    sigmasGam, musGam = gate.write_gauss_param_to_file(
        output_path,
        planePositionsV,
        saveFig=False,
        fNamePrefix="plane",
        fNameSuffix="a.mhd",
    )
else:
    print("Some data are already available for analysis")

# ----------------------------------------------------------------------------------------------------------------
# tests

# statistics
stat_file = "SimulationStatistic_" + folder + ".txt"
stats_ref = gate.read_stat_file(ref_path / stat_file)
is_ok = gate.assert_stats(stat, stats_ref, 0.10)

# energy deposition
for i in planePositionsV:
    print("\nDifference for EDEP plane " + str(i))
    mhd_gate = "plane" + str(i) + "a.mhd"
    mhd_ref = "plane" + str(i) + "a_" + folder + "-Edep.mhd"
    is_ok = (
        gate.assert_images(
            ref_path / mhd_ref,
            output_path / mhd_gate,
            stat,
            tolerance=50,
            ignore_value=0,
        )
        and is_ok
    )
    """EdepColorMap = gate.create_2D_Edep_colorMap(output_path / mhd_gate)
	img_name = 'Plane_'+str(i)+'ColorMap.png'
	EdepColorMap.savefig(output_path / img_name)
	plt.close(EdepColorMap)"""

# beam shape
sigma_file = "sigma_values.txt"
is_ok = (
    gate.compareGaussParamFromFile(
        output_path / sigma_file,
        ref_path / sigma_file,
        rel_tol=2,
        abs_tol=0.5,
        verb=True,
    )
    and is_ok
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect
import itk
import numpy as np

paths = gate.get_default_test_paths(__file__, "gate_test028_ge_nm670_spect")


def create_spect_simu(
    sim,
    paths,
    number_of_threads=1,
    activity_kBq=300,
    aa_enabled=True,
    aa_mode="SkipEnergy",
):
    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.number_of_threads = number_of_threads
    ui.check_volumes_overlap = False
    ui.random_engine = "MixMaxRng"
    ui.random_seed = 123456789

    # units
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    kBq = 1000 * Bq

    # world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_AIR"

    # spect head (debug mode = very small collimator)
    spect, crystal = gate_spect.add_ge_nm67_spect_head(
        sim, "spect", collimator_type="lehr", debug=False
    )

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [15 * cm, 15 * cm, 15 * cm]
    waterbox.material = "G4_WATER"
    waterbox.translation = [0, 0, 0]
    blue = [0, 1, 1, 1]
    waterbox.color = blue

    # physic list
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    p.enable_decay = False
    cuts = p.production_cuts
    cuts.world.gamma = 10 * mm
    cuts.world.electron = 10 * mm
    cuts.world.positron = 10 * mm
    cuts.world.proton = 10 * mm
    cuts.spect.gamma = 0.1 * mm
    cuts.spect.electron = 0.1 * mm
    cuts.spect.positron = 0.1 * mm

    # default source for tests
    # activity = 300 * kBq
    activity = activity_kBq * kBq
    beam1 = sim.add_source("GenericSource", "beam1")
    beam1.mother = waterbox.name
    beam1.particle = "gamma"
    beam1.energy.mono = 140.5 * keV
    beam1.position.type = "sphere"
    beam1.position.radius = 1 * cm
    beam1.position.translation = [0, 0, 0]
    beam1.direction.type = "iso"
    if aa_enabled:
        beam1.direction.acceptance_angle.volumes = ["spect"]
        beam1.direction.acceptance_angle.intersection_flag = True
        beam1.direction.acceptance_angle.skip_policy = aa_mode
    beam1.activity = activity / ui.number_of_threads

    beam2 = sim.add_source("GenericSource", "beam2")
    beam2.mother = waterbox.name
    beam2.particle = "gamma"
    beam2.energy.mono = 140.5 * keV
    beam2.position.type = "sphere"
    beam2.position.radius = 3 * cm
    beam2.position.translation = [18 * cm, 0, 0]
    beam2.direction.type = "iso"
    if aa_enabled:
        beam2.direction.acceptance_angle.volumes = ["spect"]
        beam2.direction.acceptance_angle.intersection_flag = True
        beam2.direction.acceptance_angle.skip_policy = aa_mode
    beam2.activity = activity / ui.number_of_threads

    beam3 = sim.add_source("GenericSource", "beam3")
    beam3.mother = waterbox.name
    beam3.particle = "gamma"
    beam3.energy.mono = 140.5 * keV
    beam3.position.type = "sphere"
    beam3.position.radius = 1 * cm
    beam3.position.translation = [0, 10 * cm, 0]
    beam3.direction.type = "iso"
    if aa_enabled:
        beam3.direction.acceptance_angle.volumes = ["spect"]
        beam3.direction.acceptance_angle.intersection_flag = True
        beam3.direction.acceptance_angle.skip_policy = aa_mode
    beam3.activity = activity / ui.number_of_threads

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    # get crystal volume by looking for the word crystal in the name
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    hc.mother = crystal.name
    hc.output = ""  # No output paths.output / 'test028.root'
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PostStepUniqueVolumeID",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]

    # singles collection
    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.mother = crystal.name
    sc.input_digi_collection = "Hits"
    sc.policy = "EnergyWinnerPosition"
    # sc.policy = 'EnergyWeightedCentroidPosition'
    sc.skip_attributes = ["KineticEnergy", "ProcessDefinedStep", "KineticEnergy"]
    sc.output = hc.output

    # EnergyWindows
    cc = sim.add_actor("DigitizerEnergyWindowsActor", "EnergyWindows")
    cc.mother = crystal.name
    cc.input_digi_collection = "Singles"
    cc.channels = [
        {"name": "scatter", "min": 114 * keV, "max": 126 * keV},
        {"name": "peak140", "min": 126 * keV, "max": 154.55 * keV},
        # {'name': 'spectrum', 'min': 0 * keV, 'max': 5000 * keV}  # should be strictly equal to 'Singles'
    ]
    cc.output = hc.output

    # projection
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    # 2D binning projection
    proj = sim.add_actor("HitsProjectionActor", "Projection")
    proj.mother = crystal.name
    # we set two times the spectrum channel to compare with Gate output
    proj.input_digi_collections = ["Singles", "scatter", "peak140", "Singles"]
    proj.spacing = [4.41806 * mm, 4.41806 * mm]
    proj.size = [128, 128]
    # proj.plane = 'XY' # not implemented yet
    proj.output = paths.output / "proj028_colli.mhd"

    # rotate spect
    cm = gate.g4_units("cm")
    psd = 6.11 * cm
    p = [0, 0, -(15 * cm + psd)]
    spect.translation, spect.rotation = gate.get_transform_orbiting(p, "y", 15)
    print("rotation 15 deg and translation = ", spect.translation)

    return spect, proj


def compare_result(output, proj, fig_name, sum_tolerance=8):
    gate.warning("Compare acceptance angle skipped particles")
    stats = output.get_actor("Stats")

    reference_ratio = 691518 / 2998895  # (23%)
    b1 = gate.get_source_zero_events(output, "beam1")
    b2 = gate.get_source_zero_events(output, "beam2")
    b3 = gate.get_source_zero_events(output, "beam3")
    print(f"Number of zeros events: {b1} {b2} {b3}")

    print(f"Number of simulated events: {stats.counts.event_count}")
    beam1 = output.get_source("beam1")
    mode = beam1.user_info.direction.acceptance_angle.skip_policy
    stats_ref = gate.read_stat_file(paths.gate_output / "stat4.txt")

    if mode == "SkipEvents":
        b1 = gate.get_source_skipped_events(output, "beam1")
        b2 = gate.get_source_skipped_events(output, "beam2")
        b3 = gate.get_source_skipped_events(output, "beam3")
        stats.counts.event_count = stats.counts.event_count + b1 + b2 + b3
        print(f"Skip Events mode, adding the skipped ones")
        print(f"Number of simulated events: {stats.counts.event_count}")
        # do not compare track in this mode
        stats.counts.track_count = stats_ref.counts.track_count

    tol = 0.3
    r1 = b1 / stats.counts.event_count
    is_ok = (r1 - reference_ratio) / reference_ratio < tol
    gate.print_test(
        is_ok,
        f"Skipped particles b1 = {b1} {r1 * 100:.2f} %  vs {reference_ratio * 100:.2f} % ",
    )

    r2 = b2 / stats.counts.event_count
    is_ok = (r2 - reference_ratio) / reference_ratio < tol
    gate.print_test(
        is_ok,
        f"Skipped particles b2 = {b2} {r2 * 100:.2f} %  vs {reference_ratio * 100:.2f} % ",
    )

    r3 = b3 / stats.counts.event_count
    is_ok = (r3 - reference_ratio) / reference_ratio < tol
    gate.print_test(
        is_ok,
        f"Skipped particles b3 = {b3} {r3 * 100:.2f} %  vs {reference_ratio * 100:.2f} % ",
    )

    # stat
    gate.warning("Compare stats")
    print(stats)
    print(f"Number of runs was {stats.counts.run_count}. Set to 1 before comparison")
    stats.counts.run_count = 1  # force to 1
    print(
        f"Number of steps was {stats.counts.step_count}, force to the same value (because of angle acceptance). "
    )
    stats.counts.step_count = stats_ref.counts.step_count  # force to id
    is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.07) and is_ok

    # read image and force change the offset to be similar to old Gate
    gate.warning("Compare projection image")
    img = itk.imread(str(paths.output / "proj028_colli.mhd"))
    spacing = np.array([proj.spacing[0], proj.spacing[1], 1])
    print("spacing", spacing)
    origin = spacing / 2.0
    origin[2] = 0.5
    spacing[2] = 1
    img.SetSpacing(spacing)
    img.SetOrigin(origin)
    itk.imwrite(img, str(paths.output / "proj028_colli_offset.mhd"))
    # There are not enough event to make a proper comparison, so the tol is very high
    is_ok = (
        gate.assert_images(
            paths.gate_output / "projection4.mhd",
            paths.output / "proj028_colli_offset.mhd",
            stats,
            tolerance=83,
            ignore_value=0,
            axis="x",
            sum_tolerance=sum_tolerance,
            fig_name=str(paths.output / fig_name),
        )
        and is_ok
    )

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4
from scipy.spatial.transform import Rotation
import pathlib

current_path = pathlib.Path(__file__).parent.resolve()
ref_path = current_path / ".." / "data" / "output_ref"
output_path = current_path / ".." / "output"

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.check_volumes_overlap = True

#  change world size
m = gate.g4_units("m")
mm = gate.g4_units("mm")
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]

# add a simple volume
airBox = sim.add_volume("Box", "AirBox")
cm = gate.g4_units("cm")
airBox.size = [30 * cm, 30 * cm, 30 * cm]
airBox.translation = [0 * cm, 0 * cm, 0 * cm]
airBox.material = "G4_AIR"
airBox.color = [0, 0, 1, 1]  # blue

# lyso material
n = g4.G4NistManager.Instance()
print(n)
elems = ["Lu"]  # , 'Yttrium', 'Silicon', 'Oxygen']
nbAtoms = [18]  # , 2, 10, 50]
gcm3 = gate.g4_units("g/cm3")
n.ConstructNewMaterialNbAtoms("LYSO", elems, nbAtoms, 7.1 * gcm3)

# repeat a box
crystal = sim.add_volume("Box", "crystal")
crystal.mother = "AirBox"
crystal.size = [1 * cm, 1 * cm, 1 * cm]
crystal.translation = None
crystal.rotation = None
crystal.material = "LYSO"
m = Rotation.identity().as_matrix()
le = [
    {"name": "crystal1", "translation": [1 * cm, 0 * cm, 0], "rotation": m},
    {"name": "crystal2", "translation": [0.2 * cm, 2 * cm, 0], "rotation": m},
    {"name": "crystal3", "translation": [-0.2 * cm, 4 * cm, 0], "rotation": m},
    {"name": "crystal4", "translation": [0, 6 * cm, 0], "rotation": m},
]
print(crystal)
print(le)
crystal.repeat = le

# WARNING:
# For large number of repetition, look test028 with RepeatParameterised volume
# (it is more efficient)

# default source for tests
source = sim.add_source("GenericSource", "Default")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
source.particle = "gamma"
source.energy.mono = 0.511 * MeV
source.position.type = "sphere"
source.position.radius = 1 * cm
source.position.translation = [0, 0, 0]
source.direction.type = "momentum"
source.direction.momentum = [0, 1, 0]
source.activity = 10000 * Bq

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# dose actor
d = sim.add_actor("DoseActor", "dose")
d.output = output_path / "test017-edep.mhd"
# d.output = ref_path / 'test017-edep-ref.mhd'
d.mother = "crystal"
d.size = [150, 150, 150]
d.spacing = [1 * mm, 1 * mm, 1 * mm]
d.translation = [5 * mm, 0 * mm, 0 * mm]
d.physical_volume_index = 0
print(
    "The Dose actor is triggered every time a hit occurs in the (logical volume) "
    '"crystal" (and any of its associated repeated physical volumes).'
)
print(
    "The Dose actor is attached to the first (repeated) crystal, it moves with its coord system."
)

# start simulation
output = sim.start()

# print results
stats = output.get_actor("Stats")
# stats.write(ref_path / 'test017-stats-ref.txt')

# tests
stats_ref = gate.read_stat_file(ref_path / "test017-stats-ref.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.04)
is_ok = (
    gate.assert_images(
        ref_path / "test017-edep-ref.mhd",
        output_path / "test017-edep.mhd",
        stats,
        tolerance=70,
    )
    and is_ok
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec
import itk
import json

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# shhhht !
gate.log.setLevel(gate.NONE)

# world
m = gate.g4_units("m")
sim.world.size = [1 * m, 1 * m, 1 * m]

# add a iec phantom
iec = gate_iec.add_phantom(sim)

# create an empty image with the size (extent) of the volume
# add one pixel margin
image = gate.create_image_with_volume_extent(sim, iec.name, spacing=[3, 3, 3], margin=1)
info = gate.get_info_from_image(image)
print(f"Image : {info.size} {info.spacing} {info.origin}")


# we need a simulation engine to voxelize a volume
# (we will reuse the engine, so it need to not be in a subprocess)
se = gate.SimulationEngine(sim, start_new_process=False)
se.initialize()

# voxelized a volume
print("Starting voxelization ...")
labels, image = gate.voxelize_volume(se, iec.name, image)
print(f"Output labels: {labels}")

# write labels
lf = str(paths.output / "test032_labels_3mm.json")
outfile = open(lf, "w")
json.dump(labels, outfile, indent=4)

# write image
f = paths.output / "test032_iec_3mm.mhd"
print(f"Write image {f}")
itk.imwrite(image, str(f))

#
# redo the same but with 1 mm spacing
#

# create an empty image with the size (extent) of the volume
# add one pixel margin
image = gate.create_image_with_volume_extent(sim, iec.name, spacing=[1, 1, 1], margin=1)
info = gate.get_info_from_image(image)
print(f"Image : {info.size} {info.spacing} {info.origin}")

# voxelized a volume
print("Starting voxelization ...")
labels, image = gate.voxelize_volume(se, iec.name, image)
print(f"Output labels: {labels}")

# write labels
lf = str(paths.output / "test032_labels.json")
outfile = open(lf, "w")
json.dump(labels, outfile, indent=4)

# write image
f = paths.output / "test032_iec.mhd"
print(f"Write image {f}")
itk.imwrite(image, str(f))

# read and compare labels
gate.warning("\nDifference labels")
ref_labels = open(paths.output_ref / "test032_labels.json").read()
ref_labels = json.loads(ref_labels)
added, removed, modified, same = gate.dict_compare(ref_labels, labels)
is_ok = len(added) == 0 and len(removed) == 0 and len(modified) == 0
gate.print_test(is_ok, f"Labels comparisons, added:    {added}")
gate.print_test(is_ok, f"Labels comparisons, removed:  {removed}")
gate.print_test(is_ok, f"Labels comparisons: modified: {modified}")

# compare images
gate.warning("\nDifference with ref image")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test032_iec.mhd", f, stats=None, tolerance=0.01
    )
    and is_ok
)

gate.test_ok(is_ok)



import opengate as gate
from opengate import log


class FilterManager:
    """
    Manage all the Filters in the simulation
    """

    def __init__(self, simulation):
        self.simulation = simulation
        self.user_info_filters = {}
        self.filters = {}

    def __str__(self):
        v = [v.name for v in self.user_info_filters.values()]
        s = f'{" ".join(v)} ({len(self.user_info_filters)})'
        return s

    def __del__(self):
        # print("del FilterManager")
        pass

    def dump(self):
        n = len(self.user_info_filters)
        s = f"Number of filters: {n}"
        for Filter in self.user_info_filters.values():
            if n > 1:
                a = "\n" + "-" * 20
            else:
                a = ""
            a += f"\n {Filter}"
            s += gate.indent(2, a)
        return s

    def get_filter(self, name):
        if name not in self.filters:
            gate.fatal(
                f"The Filter {name} is not in the current "
                f"list of Filters: {self.filters}"
            )
        return self.filters[name]

    def add_filter(self, filter_type, name):
        # check that another element with the same name does not already exist
        gate.assert_unique_element_name(self.filters, name)
        # build it
        a = gate.UserInfo("Filter", filter_type, name)
        # append to the list
        self.user_info_filters[name] = a
        # return the info
        return a

    def initialize(self):
        print("filter init")
        for ui in self.user_info_filters.values():
            print(ui)
            filter = gate.new_element(ui, self.simulation)
            log.debug(f"Filter: initialize [{ui.type_name}] {ui.name}")
            filter.Initialize(ui)
            self.filters[ui.name] = filter
            print(filter)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import numpy as np
import matplotlib.pyplot as plt

paths = gate.get_default_test_paths(__file__, "")

"""
Test the option in GenericSource to use a TAC Time Activity Curve
"""

# create the simulation
sim = gate.Simulation()

# units
m = gate.g4_units("m")
nm = gate.g4_units("nm")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
Bq = gate.g4_units("Bq")
kBq = 1e3 * Bq
sec = gate.g4_units("s")
min = gate.g4_units("min")
keV = gate.g4_units("keV")

# verbose
ui = sim.user_info
ui.visu = False
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.number_of_threads = 1
ui.random_seed = "auto"

# world size
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]
world.material = "G4_AIR"

# source of Ac225 ion
s1 = sim.add_source("GenericSource", "source")
s1.particle = "gamma"
s1.position.type = "sphere"
s1.position.radius = 1 * nm
s1.position.translation = [0, 0, 0]
s1.direction.type = "iso"
s1.energy.mono = 140 * keV

# create a decay TAC bw 2 and 5 sec
starting_activity = 100 * kBq / ui.number_of_threads
half_life = 2 * sec
times = np.linspace(1, 6, num=500, endpoint=True) * sec
decay = np.log(2) / half_life
activities = [starting_activity * np.exp(-decay * t) for t in times]
s1.tac_times = times
s1.tac_activities = activities

# stats
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

# phsp actor for timing
phsp = sim.add_actor("PhaseSpaceActor", "phsp")
phsp.attributes = ["GlobalTime"]
phsp.output = paths.output / "test052_tac.root"

# go
# ui.running_verbose_level = gate.EVENT
# (purposely a "hole" in the runtime ; warning not too large otherwise the fit fails)
sim.run_timing_intervals = [[0, 2.9 * sec], [3 * sec, 7 * sec]]
output = sim.start(start_new_process=True)

# print
stats = output.get_actor("Stats")
print(stats)

# check root
print()
gate.warning("Check root time")
root1, n1 = gate.open_root_as_np(phsp.output, "phsp")
etimes = root1["GlobalTime"] / sec
print(f"Number of events : {len(etimes)}")

# check fit (it is important to convert 'etimes' to sec otherwise
# the fit fails)
hl, xx, yy = gate.fit_exponential_decay(
    etimes, 0, 7
)  # times[0] / sec, times[-1] / sec)
tol = 0.05
hl_ref = half_life / sec
diff = abs(hl - hl_ref) / hl_ref
is_ok = b = diff < tol
diff *= 100
gate.print_test(b, f"Half life {hl_ref:.2f} sec vs {hl:.2f} sec : {diff:.2f}% ")

# plot the fit
f, ax = plt.subplots(1, 1, figsize=(25, 10))
gate.plot_hist(ax, etimes, f"Events times")
ax.plot(xx, yy, label=f"fit")
ax.legend()

fn = paths.output / "test052_tac_times.png"
plt.savefig(fn)
print(f"Plot in {fn}")

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4


class SimulationUserInfo:
    """
    This class is a simple structure that contains all user general options of a simulation.
    """

    def __init__(self, simulation):
        # keep pointer to ref
        self.simulation = simulation

        # gate (pre-run) verbose
        # A number or gate.NONE or gate.INFO or gate.DEBUG
        self._verbose_level = gate.INFO
        gate.log.setLevel(self._verbose_level)

        # gate verbose during running
        self.running_verbose_level = 0

        # Geant4 verbose
        # For an unknown reason, when verbose_level == 0, there are some
        # additional print after the G4RunManager destructor. So we default at 1
        self.g4_verbose_level = 1
        self.g4_verbose = False

        # visualisation (qt|vrml)
        self.visu = False
        self.visu_type = "qt"  # choice: "qt", "vrml" or "gdml"
        self.visu_filename = "visu.wrl"
        self.visu_verbose = False
        self.visu_commands = gate.read_mac_file_to_commands("default_visu_commands.mac")
        self.visu_commands_vrml = gate.read_mac_file_to_commands(
            "default_visu_commands_vrml.mac"
        )
        self.visu_commands_gdml = gate.read_mac_file_to_commands(
            "default_visu_commands_gdml.mac"
        )

        # check volume overlap once constructed
        self.check_volumes_overlap = True

        # multi-threading
        self.number_of_threads = 1
        self.force_multithread_mode = False

        # random engine
        # MixMaxRng seems recommended for MultiThread
        self.random_engine = "MixMaxRng"  # 'MersenneTwister'
        self.random_seed = "auto"

    @property
    def verbose_level(self):
        return self._verbose_level

    @verbose_level.setter
    def verbose_level(self, value):
        gate.log.setLevel(value)
        self._verbose_level = value

    def __del__(self):
        pass

    def __str__(self):
        if self.number_of_threads == 1 and not self.force_multithread_mode:
            g = g4.GateInfo.get_G4MULTITHREADED()
            t = "no"
            if g:
                t += " (but available: G4 was compiled with MT)"
            else:
                t += " (not available, G4 was not compiled with MT)"
        else:
            t = f"{self.number_of_threads} threads"
        s = (
            f"Verbose         : {self.verbose_level}\n"
            f"Running verbose : {self.running_verbose_level}\n"
            f"Geant4 verbose  : {self.g4_verbose}, level = {self.g4_verbose_level}\n"
            f"Visualisation   : {self.visu}, verbose level = {self.g4_verbose_level}\n"
            f"Visutype        : {self.visu_type}\n"
            f"Check overlap   : {self.check_volumes_overlap}\n"
            f"Multithreading  : {t}\n"
            f"Random engine   : {self.random_engine}, seed = {self.random_seed}"
        )
        return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import matplotlib.pyplot as plt
import sys

paths = gate.get_default_test_paths(__file__, "test050_let_actor_letd")

ref_path = paths.output_ref


# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_seed = 12345678910


numPartSimTest = 20000
numPartSimRef = 1e5


# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
km = gate.g4_units("km")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq


#  change world size
world = sim.world
world.size = [600 * cm, 500 * cm, 500 * cm]
# world.material = "Vacuum"

# waterbox
phantom = sim.add_volume("Box", "phantom")
phantom.size = [10 * cm, 10 * cm, 10 * cm]
phantom.translation = [-5 * cm, 0, 0]
phantom.material = "G4_WATER"
phantom.color = [0, 0, 1, 1]


test_material_name = "G4_WATER"
phantom_off = sim.add_volume("Box", "phantom_off")
phantom_off.mother = phantom.name
phantom_off.size = [100 * mm, 60 * mm, 60 * mm]
phantom_off.translation = [0 * mm, 0 * mm, 0 * mm]
phantom_off.material = test_material_name
phantom_off.color = [0, 0, 1, 1]

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BIC_EMZ"
# sim.set_cut("world", "all", 1000 * km)
# FIXME need SetMaxStepSizeInRegion ActivateStepLimiter

# default source for tests
source = sim.add_source("GenericSource", "mysource")
source.energy.mono = 80 * MeV
# source.energy.type = 'gauss'
# source.energy.sigma_gauss = 1 * MeV
source.particle = "proton"
source.position.type = "disc"
source.position.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
source.position.radius = 4 * mm
source.position.translation = [0, 0, 0]
source.direction.type = "momentum"
source.direction.momentum = [-1, 0, 0]
# print(dir(source.energy))
source.n = numPartSimTest
# source.activity = 100 * kBq


# filter : keep proton
# f = sim.add_filter("ParticleFilter", "f")
# f.particle = "proton"


size = [50, 1, 1]
spacing = [2.0 * mm, 60.0 * mm, 60.0 * mm]


doseActorName_IDD_d = "IDD_d"
doseIDD = sim.add_actor("DoseActor", doseActorName_IDD_d)
doseIDD.output = paths.output / ("test050-" + doseActorName_IDD_d + ".mhd")
doseIDD.mother = phantom_off.name
doseIDD.size = size
doseIDD.spacing = spacing
doseIDD.hit_type = "random"
doseIDD.gray = False


LETActorName_IDD_d = "LETActorOG_d"
LETActor_IDD_d = sim.add_actor("LETActor", LETActorName_IDD_d)
LETActor_IDD_d.output = paths.output / ("test050-" + LETActorName_IDD_d + ".mhd")
LETActor_IDD_d.mother = phantom_off.name
LETActor_IDD_d.size = size
LETActor_IDD_d.spacing = spacing
LETActor_IDD_d.hit_type = "random"
LETActor_IDD_d.separate_output = True
## both lines do the same thing,
setattr(
    LETActor_IDD_d, "dose_average", True
)  ## usesful for looping over several options
# LETActor_IDD_d.track_average = True ## same as above line


LETActorName_IDD_t = "LETActorOG_t"
LETActor_IDD_t = sim.add_actor("LETActor", LETActorName_IDD_t)
LETActor_IDD_t.output = paths.output / ("test050-" + LETActorName_IDD_t + ".mhd")
LETActor_IDD_t.mother = phantom_off.name
LETActor_IDD_t.size = size
LETActor_IDD_t.spacing = spacing
LETActor_IDD_t.hit_type = "random"
LETActor_IDD_t.track_average = True


LETActorName_IDD_d2w = "LETActorOG_d2w"
LETActor_IDD_d2w = sim.add_actor("LETActor", LETActorName_IDD_d2w)
LETActor_IDD_d2w.output = paths.output / ("test050-" + LETActorName_IDD_d2w + ".mhd")
LETActor_IDD_d2w.mother = phantom_off.name
LETActor_IDD_d2w.size = size
LETActor_IDD_d2w.spacing = spacing
LETActor_IDD_d2w.hit_type = "random"
LETActor_IDD_d2w.other_material = "G4_WATER"
setattr(LETActor_IDD_d2w, "dose_average", True)


fName_ref_IDD = "IDD__Proton_Energy1MeVu_RiFiout-Edep.mhd"
print(paths)
# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "stats")
s.track_types_flag = True
# s.filters.append(f)

print("Filters: ", sim.filter_manager)
print(sim.filter_manager.dump())

# start simulation
sim.n = 10
output = sim.start()


ref_path = paths.output_ref
# paths.gate_output

# print results at the end
stat = output.get_actor("stats")
print(stat)

# ----------------------------------------------------------------------------------------------------------------
# tests
print()
# gate.warning("Tests stats file")
# stats_ref = gate.read_stat_file(paths.gate_output / "stats.txt")
# is_ok = gate.assert_stats(stat, stats_ref, 0.14)

LETActorFPath_doseAveraged = output.get_actor(LETActorName_IDD_d).user_info.output
LETActorFPath_trackAveraged = output.get_actor(LETActorName_IDD_t).user_info.output

fNameIDD = "test050_IDD__Proton_Energy1MeVu_RiFiout-Edep.mhd"
"""
is_ok = gate.assert_images(
    ref_path / fNameIDD,
    doseIDD.output,
    stat,
    tolerance=100,
    ignore_value=0,
    axis="x",
    scaleImageValuesFactor=numPartSimRef / numPartSimTest,
)
"""

is_ok = gate.assert_filtered_imagesprofile1D(
    ref_filter_filename1=ref_path / fNameIDD,
    ref_filename1=ref_path / "test050_LET1D_noFilter__PrimaryProton-doseAveraged.mhd",
    filename2=LETActor_IDD_d.output,
    tolerance=15,
    plt_ylim=[0, 25],
)

is_ok = (
    gate.assert_filtered_imagesprofile1D(
        ref_filter_filename1=ref_path / fNameIDD,
        ref_filename1=ref_path
        / "test050_LET1D_noFilter__PrimaryProton-trackAveraged.mhd",
        filename2=LETActor_IDD_t.output,
        tolerance=8,
        plt_ylim=[0, 18],
    )
    and is_ok
)


gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test037_pet_hits_singles_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test037_pet")

# test version
v = "2_2"

# create the simulation
sim = gate.Simulation()
crystal = create_pet_simulation(sim, paths)
module = sim.get_volume_user_info("pet_module")
die = sim.get_volume_user_info("pet_die")
stack = sim.get_volume_user_info("pet_stack")

# digitizer hits
hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
hc.mother = crystal.name
hc.output = paths.output / f"test037_test{v}.root"
hc.attributes = [
    "PostPosition",
    "TotalEnergyDeposit",
    "PreStepUniqueVolumeID",
    "GlobalTime",
]

# Readout (not need for adder)
sc = sim.add_actor("DigitizerReadoutActor", "Singles2_1")
sc.output = paths.output / f"test037_test{v}.root"
sc.input_digi_collection = "Hits"
sc.group_volume = stack.name  # should be depth=1 in Gate
sc.discretize_volume = crystal.name
sc.policy = "EnergyWeightedCentroidPosition"

# Readout: another one, with different option (in the same output file)
sc = sim.add_actor("DigitizerReadoutActor", "Singles2_2")
sc.output = paths.output / f"test037_test{v}.root"
sc.input_digi_collection = "Hits"
sc.group_volume = crystal.name  # should be depth=4 in Gate
sc.discretize_volume = crystal.name
sc.policy = "EnergyWeightedCentroidPosition"

# timing
sec = gate.g4_units("second")
sim.run_timing_intervals = [[0, 0.00005 * sec]]

# start simulation
output = sim.start()

# print results
stats = output.get_actor("Stats")
print(stats)

# ----------------------------------------------------------------------------------------------------------

# check stats
print()
gate.warning(f"Check stats")
# p = paths.gate / "output_test1"
p = paths.gate / "output"
stats_ref = gate.read_stat_file(p / f"stats{v}.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.025)

# check root hits
hc = output.get_actor("Hits").user_info
f = p / f"output{v}.root"
is_ok = check_root_hits(paths, v, f, hc.output) and is_ok

# check root singles
sc = output.get_actor("Singles2_1").user_info
f = p / f"output2_1.root"
is_ok = check_root_singles(paths, "2_1", f, sc.output, sc.name) and is_ok

# check root singles
sc = output.get_actor("Singles2_2").user_info
f = p / f"output2_2.root"
is_ok = check_root_singles(paths, "2_2", f, sc.output, sc.name) and is_ok

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantom_nema_iec_body as gate_iec
import gatetools.phsp as phsp
import uproot
import numpy as np
import matplotlib.pyplot as plt
import os

paths = gate.get_default_test_paths(__file__, "")
paths.output_ref = paths.output_ref / "test040_ref"

# create the simulation
sim = gate.Simulation()

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
cm3 = gate.g4_units("cm3")
keV = gate.g4_units("keV")
MeV = gate.g4_units("MeV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
BqmL = Bq / cm3
sec = gate.g4_units("second")
deg = gate.g4_units("deg")
kBq = 1000 * Bq
MBq = 1000 * kBq

# main parameters
ui = sim.user_info
ui.check_volumes_overlap = True
ui.number_of_threads = 1
ui.random_seed = 123456
# ui.running_verbose_level = gate.EVENT
# ui.g4_verbose = True
ac = 5e3 * BqmL / ui.number_of_threads
ui.visu = False
colli_flag = not ui.visu
if ui.visu:
    ac = 1 * BqmL
    ui.number_of_threads = 1

# world size
world = sim.world
world.size = [1.5 * m, 1.5 * m, 1.5 * m]
world.material = "G4_AIR"

# test phase space to check with reference
phsp_sphere_surface = sim.add_volume("Sphere", "phase_space_sphere")
phsp_sphere_surface.rmin = 215 * mm
phsp_sphere_surface.rmax = 216 * mm
phsp_sphere_surface.color = [1, 1, 1, 1]
phsp_sphere_surface.material = "G4_AIR"

# physic list
sim.set_cut("world", "all", 1 * mm)

# activity parameters
spheres_diam = [10, 13, 17, 22, 28, 37]
spheres_activity_concentration = [ac * 6, ac * 5, ac * 4, ac * 3, ac * 2, ac]

# initialisation for conditional
spheres_radius = [x / 2.0 for x in spheres_diam]
spheres_centers, spheres_volumes = gate_iec.get_default_sphere_centers_and_volumes()
spheres_activity_ratio = []
spheres_activity = []
for diam, ac, volume, center in zip(
    spheres_diam, spheres_activity_concentration, spheres_volumes, spheres_centers
):
    activity = ac * volume
    print(
        f"Sphere {diam}: {str(center):<30} {volume / cm3:7.3f} cm3 "
        f"{activity / Bq:7.0f} Bq  {ac / BqmL:7.1f} BqmL"
    )
    spheres_activity.append(activity)

total_activity = sum(spheres_activity)
print(f"Total activity {total_activity / Bq:.0f} Bq")
for activity in spheres_activity:
    spheres_activity_ratio.append(activity / total_activity)
print("Activity ratio ", spheres_activity_ratio, sum(spheres_activity_ratio))

# will store all conditional info (position, direction)
all_cond = None

# unique (reproducible) random generator
rs = gate.get_rnd_seed(123456)


def gen_cond(n):
    n_samples = gate_iec.get_n_samples_from_ratio(n, spheres_activity_ratio)
    # (it is very important to shuffle when several spheres to avoid time artifact)
    cond = gate_iec.generate_pos_spheres(
        spheres_centers, spheres_radius, n_samples, shuffle=True, rs=rs
    )

    # we keep all conditions for the test (not needed in normal simulation)
    global all_cond
    if all_cond is None:
        all_cond = cond
    else:
        all_cond = np.column_stack((all_cond, cond))

    return cond


# GAN source
gsource = sim.add_source("GANPairsSource", "gaga")
gsource.particle = "gamma"
# no phantom, we consider attached to the world at origin
gsource.activity = total_activity
gsource.pth_filename = paths.data / "pth120_test9221_GP_0GP_10.0_100000.pth"
gsource.position_keys = ["X1", "Y1", "Z1", "X2", "Y2", "Z2"]
gsource.direction_keys = ["dX1", "dY1", "dZ1", "dX2", "dY2", "dZ2"]
gsource.energy_key = ["E1", "E2"]
gsource.time_key = ["t1", "t2"]
# time is added to the simulation time
gsource.relative_timing = True
gsource.weight_key = None
# particle are move backward with 10 cm
gsource.backward_distance = 10 * cm
gsource.backward_force = True
# if the kinetic E is below this threshold, we set it to 0
gsource.energy_min_threshold = 0.1 * keV
gsource.energy_max_threshold = 1 * MeV
gsource.skip_policy = "ZeroEnergy"
gsource.batch_size = 1e5
gsource.verbose_generator = True
# set the generator and the condition generator
gsource.generator = gate.GANSourceConditionalPairsGenerator(gsource, 210 * mm, gen_cond)

# add stat actor
stat = sim.add_actor("SimulationStatisticsActor", "Stats")
stat.output = paths.output / "test040_gan_stats.txt"

# phsp actor
phsp_actor = sim.add_actor("PhaseSpaceActor", "phsp")
phsp_actor.mother = phsp_sphere_surface.name
phsp_actor.attributes = [
    "KineticEnergy",
    "PrePosition",
    "PreDirection",
    "GlobalTime",
    "EventPosition",
    "EventDirection",
    "TimeFromBeginOfEvent",
    "EventKineticEnergy",
]
phsp_actor.output = paths.output / "test040_gan_phsp.root"
f = sim.add_filter("ParticleFilter", "f")
f.particle = "gamma"
phsp_actor.filters.append(f)
f = sim.add_filter("KineticEnergyFilter", "f")
f.energy_min = 100 * keV
phsp_actor.filters.append(f)

# ----------------------------------------------------------------------------------------------
# go (cannot be spawn in another process)
# ui.running_verbose_level = gate.EVENT
output = sim.start(False)

# ----------------------------------------------------------------------------------------------
# print stats
print()
gate.warning(f"Check stats")
if ui.number_of_threads == 1:
    s = output.get_source("gaga")
else:
    s = output.get_source_MT("gaga", 0)
print(f"Source, nb of skipped particles : {s.fTotalSkippedEvents}")
b = gate.get_source_skipped_events(output, gsource.name)
print(f"Source, nb of skipped particles (check) : {b}")

print(f"Source, nb of zerosE particles : {s.fTotalZeroEvents}")
b = gate.get_source_zero_events(output, gsource.name)
print(f"Source, nb of zerosE particles (check) : {b}")


stats = output.get_actor("Stats")
print(stats)
stats_ref = gate.read_stat_file(paths.output_ref / "test040_ref_stats.txt")
r = (
    stats_ref.counts.step_count - stats.counts.step_count
) / stats_ref.counts.step_count
print(f"!!! Steps cannot be compared => was {stats.counts.step_count}, {r:.2f}%")
stats.counts.step_count = stats_ref.counts.step_count
r = (
    stats_ref.counts.track_count - stats.counts.track_count
) / stats_ref.counts.track_count
print(f"!!! Tracks cannot be compared => was {stats.counts.track_count}, {r:.2f}%")
stats.counts.track_count = stats_ref.counts.track_count
is_ok = gate.assert_stats(stats, stats_ref, 0.10)

# save conditional for checking with reference cond
keys = ["EventPosition_X", "EventPosition_Y", "EventPosition_Z"]
phsp.save_npy(paths.output / "test040_gan_phsp_cond.npy", all_cond, keys)

# ----------------------------------------------------------------------------------------------
# compare conditional
# less particle in the ref because conditional data are stored
# when exit (not absorbed)
print()
gate.warning(f"Check conditions (position, direction)")
root_ref = (
    paths.output / "test040_ref_phsp.root"
)  # look root generated by previous 40_aref test
hits1 = uproot.open(root_ref)
branch = hits1.keys()[0]
print("Branch name:", branch)
hits1 = hits1[branch]
names = [k for k in hits1.keys()]
hits1_n = hits1.num_entries
hits1 = hits1.arrays(library="numpy")

# in the ref phsp, EventID is not unique (pairs of gamma), we set them unique
# to be compared to the generated cond
event_id = hits1["EventID"]
print("Nb of event (non unique)", event_id.shape)
u, indices = np.unique(event_id, return_index=True)
for k in hits1:
    hits1[k] = hits1[k][indices]
event_id = hits1["EventID"]
print("Nb of event (unique)", event_id.shape)

root_gan = paths.output / "test040_gan_phsp_cond.npy"
hits2, hits2_keys, hits2_n = phsp.load(root_gan)
tols = [10.0] * len(keys)
tols[keys.index("EventPosition_X")] = 0.15
tols[keys.index("EventPosition_Y")] = 0.15
tols[keys.index("EventPosition_Z")] = 0.15
scalings = [1] * len(keys)
is_ok = (
    gate.compare_trees(
        hits1,
        list(hits1.keys()),
        hits2,
        list(hits2_keys),
        keys,
        keys,
        tols,
        scalings,
        scalings,
        True,
    )
    and is_ok
)
# figure
img_filename = paths.output / "test040_cond.png"
plt.suptitle(
    f"Values: ref {os.path.basename(root_ref)} {os.path.basename(root_gan)} "
    f"-> {hits1_n} vs {hits2_n}"
)
plt.savefig(img_filename)
print(f"Figure in {img_filename}")

# ----------------------------------------------------------------------------------------------
# compare output phsp
# absorbed events are when E == 0
print()
gate.warning(
    f"Check output phsp (the pth is not perfect, but it should be sufficient for tests"
)
print("Warning: in the ref phsp, we remove when E==0 (absorbed events)")

ref_file = paths.output / "test040_ref_phsp.root"
hits1 = uproot.open(root_ref)
branch = hits1.keys()[0]
print("Branch name:", branch)
hits1 = hits1[branch]
names = [k for k in hits1.keys()]
hits1_n = hits1.num_entries
hits1 = hits1.arrays(library="numpy")

# in the ref, remove when E == 0
ke = hits1["KineticEnergy"]
print("Nb of event (all E)", ke.shape)
mask = ke > 0
for k in hits1:
    hits1[k] = hits1[k][mask]
ke = hits1["KineticEnergy"]
print("Nb of event (non E==0)", ke.shape)

hc_file = phsp_actor.output
hits2, hits2_keys, hits2_n = phsp.load(hc_file)

checked_keys = [
    "GlobalTime",
    "KineticEnergy",
    "PrePosition_X",
    "PrePosition_Y",
    "PrePosition_Z",
    "PreDirection_X",
    "PreDirection_Y",
    "PreDirection_Z",
]
scalings = [1.0] * len(checked_keys)
scalings[checked_keys.index("GlobalTime")] = 1e-9  # time in ns
tols = [10.0] * len(checked_keys)
tols[checked_keys.index("GlobalTime")] = 0.003
tols[checked_keys.index("KineticEnergy")] = 0.01
tols[checked_keys.index("PrePosition_X")] = 2
tols[checked_keys.index("PrePosition_Y")] = 1
tols[checked_keys.index("PrePosition_Z")] = 1
tols[checked_keys.index("PreDirection_X")] = 0.02
tols[checked_keys.index("PreDirection_Y")] = 0.02
tols[checked_keys.index("PreDirection_Z")] = 0.02
is_ok = (
    gate.compare_trees(
        hits1,
        list(hits1.keys()),
        hits2,
        list(hits2_keys),
        checked_keys,
        checked_keys,
        tols,
        scalings,
        scalings,
        True,
    )
    and is_ok
)
# figure
img_filename = paths.output / "test040_phsp.png"
plt.suptitle(
    f"Values: ref {os.path.basename(root_ref)} {os.path.basename(root_gan)} "
    f"-> {hits1_n} vs {hits2_n}"
)
plt.savefig(img_filename)
print(f"Figure in {img_filename}")

# is_ok = gate.compare_root3(ref_file, hc_file, "phsp", "phsp",
#                          checked_keys, checked_keys, tols, scalings, scalings,
#                          paths.output / 'test040_phsp.png') and is_ok

# ----------------------------------------------------------------------------------------------

# this is the end, my friend
gate.delete_run_manager_if_needed(sim)
gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test028_ge_nm670_spect_2_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test028_ge_nm670_spect")

# create the simulation
sim = gate.Simulation()

# main description
spect = create_spect_simu(sim, paths, number_of_threads=2)

# change the digitizer to add blurring between the adder and the energy window
mm = gate.g4_units("mm")
hc = sim.get_actor_user_info("Hits")
sc = sim.get_actor_user_info("Singles")
cc = sim.get_actor_user_info("EnergyWindows")

bc = sim.add_actor("DigitizerSpatialBlurringActor", "Singles_blur")
bc.output = sc.output
bc.mother = "spect_crystal"  # important !
bc.input_digi_collection = sc.name
bc.blur_attribute = "PostPosition"
r = 20 * mm
bc.blur_fwhm = 20 * mm  #  [r, r, r]
bc.keep_in_solid_limits = True

# input of EnergyWindows is the blur one
cc.input_digi_collection = "Singles_blur"

# we modify the priority to have the correct actors order
hc.priority = 90
sc.priority = 91
cc.priority = 93
bc.priority = 92

proj = test_add_proj(sim, paths)

# rotate spect
cm = gate.g4_units("cm")
psd = 6.11 * cm
p = [0, 0, -(20 * cm + psd)]
spect.translation, spect.rotation = gate.get_transform_orbiting(p, "y", -15)

sec = gate.g4_units("second")
sim.run_timing_intervals = [[1 * sec, 2 * sec]]

print(sim.user_info)
output = sim.start()

# print stats
stats = output.get_actor("Stats")
print(stats)

# check singles
print()
version = "3_blur"
gate.warning("Compare singles")
gate_file = paths.gate_output / f"hits{version}.root"
hc_file = output.get_actor("Singles_blur").user_info.output
checked_keys = [
    {"k1": "globalPosX", "k2": "PostPosition_X", "tol": 1.8, "scaling": 1},
    {"k1": "globalPosY", "k2": "PostPosition_Y", "tol": 1.3, "scaling": 1},
    {"k1": "globalPosZ", "k2": "PostPosition_Z", "tol": 0.2, "scaling": 1},
    {"k1": "energy", "k2": "TotalEnergyDeposit", "tol": 0.001, "scaling": 1},
]
is_ok = gate.compare_root2(
    gate_file,
    hc_file,
    "Singles",
    "Singles_blur",
    checked_keys,
    paths.output / f"test028_{version}_singles.png",
)

# check projection
proj = output.get_actor("Projection")
is_ok = test_spect_proj(output, paths, proj, version="3_blur") and is_ok

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4
import itk
import numpy as np


class ImageVolume(gate.VolumeBase):
    """
    Store information about a voxelized volume
    """

    type_name = "Image"

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        user_info.image = None
        user_info.material = "G4_AIR"
        user_info.voxel_materials = [[None, "G4_AIR"]]
        user_info.dump_label_image = None

    def __init__(self, user_info):
        super().__init__(user_info)
        self.g4_physical_z = None
        self.g4_logical_z = None
        self.g4_solid_z = None
        self.g4_physical_x = None
        self.g4_logical_x = None
        self.g4_solid_x = None
        self.g4_physical_y = None
        self.g4_logical_y = None
        self.g4_solid_y = None
        # the (itk) image
        self.image = None
        # the list of regions
        self.g4_regions = []

    def __del__(self):
        pass

    def construct(self, volume_engine):
        self.volume_engine = volume_engine
        # read image
        self.image = itk.imread(gate.check_filename_type(self.user_info.image))
        size_pix = np.array(itk.size(self.image)).astype(int)
        spacing = np.array(self.image.GetSpacing())
        size_mm = size_pix * spacing

        # shorter coding
        name = self.user_info.name
        hsize_mm = size_mm / 2.0
        hspacing = spacing / 2.0

        # build the bounding box volume
        self.g4_solid = g4.G4Box(name, hsize_mm[0], hsize_mm[1], hsize_mm[2])
        def_mat = volume_engine.find_or_build_material(self.user_info.material)
        self.g4_logical_volume = g4.G4LogicalVolume(self.g4_solid, def_mat, name)

        # param Y
        self.g4_solid_y = g4.G4Box(name + "_Y", hsize_mm[0], hspacing[1], hsize_mm[2])
        self.g4_logical_y = g4.G4LogicalVolume(
            self.g4_solid_y, def_mat, name + "_log_Y"
        )
        self.g4_physical_y = g4.G4PVReplica(
            name + "_Y",
            self.g4_logical_y,
            self.g4_logical_volume,
            g4.EAxis.kYAxis,
            size_pix[1],  # nReplicas
            spacing[1],  # width
            0.0,
        )  # offset

        # param X
        self.g4_solid_x = g4.G4Box(name + "_X", hspacing[0], hspacing[1], hsize_mm[2])
        self.g4_logical_x = g4.G4LogicalVolume(
            self.g4_solid_x, def_mat, name + "_log_X"
        )
        self.g4_physical_x = g4.G4PVReplica(
            name + "_X",
            self.g4_logical_x,
            self.g4_logical_y,
            g4.EAxis.kXAxis,
            size_pix[0],
            spacing[0],
            0.0,
        )

        # param Z
        self.g4_solid_z = g4.G4Box(name + "_Z", hspacing[0], hspacing[1], hspacing[2])
        self.g4_logical_z = g4.G4LogicalVolume(
            self.g4_solid_z, def_mat, name + "_log_Z"
        )
        self.initialize_image_parameterisation()
        self.g4_physical_z = g4.G4PVParameterised(
            name + "_Z",
            self.g4_logical_z,
            self.g4_logical_x,
            g4.EAxis.kZAxis,  # g4.EAxis.kUndefined, ## FIXME ?
            size_pix[2],
            self.g4_voxel_param,
            False,
        )  # overlaps checking

        # find the mother's logical volume
        vol = self.user_info
        if vol.mother:
            st = g4.G4LogicalVolumeStore.GetInstance()
            mother_logical = st.GetVolume(vol.mother, False)
        else:
            mother_logical = None

        # consider the 3D transform -> helpers_transform.
        transform = gate.get_vol_g4_transform(vol)
        self.g4_physical_volume = g4.G4PVPlacement(
            transform,
            self.g4_logical_volume,  # logical volume
            vol.name,  # volume name
            mother_logical,  # mother volume or None if World
            False,  # no boolean operation
            0,  # copy number
            True,
        )  # overlaps checking

        # construct region
        # not clear -> should we create region for all other LV ?
        # (seg fault if region for g4_logical_z)
        self.add_region(self.g4_logical_volume)

    def add_region(self, lv):
        name = lv.GetName()
        rs = g4.G4RegionStore.GetInstance()
        r = rs.FindOrCreateRegion(name)
        self.g4_regions.append(r)
        lv.SetRegion(r)
        r.AddRootLogicalVolume(lv, True)

    def initialize_image_parameterisation(self):
        """
        From the input image, a label image is computed with each label
        associated with a material.
        The label image is initialized with label 0, corresponding to the first material
        Correspondence from voxel value to material is given by a list of interval [min_value, max_value, material_name]
        all pixels with values between min (included) and max (non included)
        will be associated with the given material
        """
        self.g4_voxel_param = g4.GateImageNestedParameterisation()
        # create image with same size
        info = gate.read_image_info(str(self.user_info.image))
        self.py_image = gate.create_3d_image(
            info.size, info.spacing, pixel_type="unsigned short", fill_value=0
        )

        # sort intervals of voxels_values <-> materials
        mat = self.user_info.voxel_materials
        interval_values_inf = [row[0] for row in mat]
        interval_values_sup = [row[1] for row in mat]
        interval_materials = [row[2] for row in mat]
        indexes = np.argsort(interval_values_inf)
        interval_values_inf = list(np.array(interval_values_inf)[indexes])
        interval_values_sup = list(np.array(interval_values_sup)[indexes])
        interval_materials = list(np.array(interval_materials)[indexes])

        # build the material
        for m in interval_materials:
            self.volume_engine.find_or_build_material(m)

        # compute list of labels and material
        self.final_materials = []
        # the image is initialized with the label zero, the first material
        self.final_materials.append(self.user_info.material)

        # convert interval to material id
        input = itk.array_view_from_image(self.image)
        output = itk.array_view_from_image(self.py_image)
        # the final list of materials is packed (same label even if
        # there are several intervals with the same material)
        self.final_materials = []
        for inf, sup, m in zip(
            interval_values_inf, interval_values_sup, interval_materials
        ):
            if m in self.final_materials:
                l = self.final_materials.index(m)
            else:
                self.final_materials.append(m)
                l = len(self.final_materials) - 1
            output[(input >= inf) & (input < sup)] = l

        # dump label image ?
        if self.user_info.dump_label_image:
            self.py_image.SetOrigin(info.origin)
            itk.imwrite(self.py_image, str(self.user_info.dump_label_image))

        # compute image origin
        size_pix = np.array(itk.size(self.py_image))
        spacing = np.array(self.py_image.GetSpacing())
        orig = -(size_pix * spacing) / 2.0 + spacing / 2.0
        self.py_image.SetOrigin(orig)

        # send image to cpp size
        gate.update_image_py_to_cpp(
            self.py_image, self.g4_voxel_param.cpp_edep_image, True
        )

        # initialize parametrisation
        self.g4_voxel_param.initialize_image()
        self.g4_voxel_param.initialize_material(self.final_materials)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect
from test028_ge_nm670_spect_2_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test028_ge_nm670_spect")

# create the simulation
sim = gate.Simulation()

# main description
create_spect_simu(sim, paths, 3)

# go
output = sim.start()

# check
is_ok = test_spect_hits(output, paths)

gate.test_ok(is_ok)



import opengate_core as g4
import itk
import numpy as np
import opengate as gate
from scipy.spatial.transform import Rotation


class LETActor(g4.GateLETActor, gate.ActorBase):
    """
    LETActor: compute a 3D edep/dose map for deposited
    energy/absorbed dose in the attached volume

    The dose map is parameterized with:
        - size (number of voxels)
        - spacing (voxel size)
        - translation (according to the coordinate system of the "attachedTo" volume)
        - no rotation

    Position:
    - by default: centered according to the "attachedTo" volume center
    - if the attachedTo volume is an Image AND the option "img_coord_system" is True:
        the origin of the attachedTo image is used for the output dose.
        Hence, the dose can be superimposed with the attachedTo volume

    Options
        - LETd only for the moment
        - later: LETt, Q, fluence ...

    """

    type_name = "LETActor"

    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        # required user info, default values
        mm = gate.g4_units("mm")
        user_info.size = [10, 10, 10]
        user_info.spacing = [1 * mm, 1 * mm, 1 * mm]
        user_info.output = "LETActor.mhd"  # FIXME change to 'output' ?
        user_info.translation = [0, 0, 0]
        user_info.img_coord_system = None
        user_info.output_origin = None
        user_info.physical_volume_index = None
        user_info.hit_type = "random"

        ## Settings for LET averaging
        user_info.dose_average = False
        user_info.track_average = False
        user_info.let_to_other_material = False
        user_info.let_to_water = False
        user_info.other_material = ""
        user_info.separate_output = False

    def __init__(self, user_info):
        ## TODO: why not super? what would happen?
        gate.ActorBase.__init__(self, user_info)
        g4.GateLETActor.__init__(self, user_info.__dict__)
        # attached physical volume (at init)
        self.g4_phys_vol = None
        # default image (py side)
        self.py_numerator_image = None
        self.py_denominator_image = None
        self.py_output_image = None

        # internal states
        self.img_origin_during_run = None
        self.first_run = None
        self.output_origin = None

    def __str__(self):
        u = self.user_info
        s = f'LETActor "{u.name}": dim={u.size} spacing={u.spacing} {u.output} tr={u.translation}'
        return s

    def __getstate__(self):
        # superclass getstate
        gate.ActorBase.__getstate__(self)
        # do not pickle itk images
        self.py_numerator_image = None
        self.py_denominator_image = None
        self.py_output_image = None
        return self.__dict__

    def initialize(self, volume_engine=None):
        """
        At the start of the run, the image is centered according to the coordinate system of
        the mother volume. This function computes the correct origin = center + translation.
        Note that there is a half-pixel shift to align according to the center of the pixel,
        like in ITK.
        """
        super().initialize(volume_engine)
        # create itk image (py side)
        size = np.array(self.user_info.size)
        spacing = np.array(self.user_info.spacing)
        self.py_numerator_image = gate.create_3d_image(size, spacing, "double")
        # TODO remove code
        # self.py_denominator_image = gate.create_3d_image(size, spacing)
        # self.py_output_image = gate.create_3d_image(size, spacing)
        # compute the center, using translation and half pixel spacing
        self.img_origin_during_run = (
            -size * spacing / 2.0 + spacing / 2.0 + self.user_info.translation
        )
        # for initialization during the first run
        self.first_run = True

        if self.user_info.dose_average == self.user_info.track_average:
            gate.fatal(
                f"Ambiguous to enable dose and track averaging: \ndose_average: {self.user_info.dose_average} \ntrack_average: { self.user_info.track_average} \nOnly one option can and must be set to True"
            )

        if self.user_info.other_material:
            self.user_info.let_to_other_material = True
        if self.user_info.let_to_other_material and not self.user_info.other_material:
            gate.fatal(
                f"let_to_other_material enabled, but other_material not set: {self.user_info.other_material}"
            )
        if self.user_info.let_to_water:
            self.user_info.other_material = "G4_WATER"

    def StartSimulationAction(self):
        # init the origin and direction according to the physical volume
        # (will be updated in the BeginOfRun)
        try:
            self.g4_phys_vol = gate.get_physical_volume(
                self.volume_engine,
                self.user_info.mother,
                self.user_info.physical_volume_index,
            )
        except:
            gate.fatal(f"Error in the LETActor {self.user_info.name}")
        gate.attach_image_to_physical_volume(
            self.g4_phys_vol.GetName(),
            self.py_numerator_image,
            self.user_info.translation,
        )

        # Set the real physical volume name
        self.fPhysicalVolumeName = str(self.g4_phys_vol.GetName())

        # FIXME for multiple run and motion
        if not self.first_run:
            gate.warning(f"Not implemented yet: LETActor with several runs")
        # send itk image to cpp side, copy data only the first run.
        gate.update_image_py_to_cpp(
            self.py_numerator_image, self.cpp_numerator_image, self.first_run
        )

        # TODO
        self.py_denominator_image = gate.create_image_like(
            self.py_numerator_image, pixel_type="double"
        )
        gate.update_image_py_to_cpp(
            self.py_denominator_image, self.cpp_denominator_image, self.first_run
        )

        self.py_output_image = gate.create_image_like(self.py_numerator_image)

        # now, indicate the next run will not be the first
        self.first_run = False

        # If attached to a voxelized volume, we may want to use its coord system.
        # So, we compute in advance what will be the final origin of the dose map
        vol_name = self.user_info.mother
        vol_type = self.simulation.get_volume_user_info(vol_name).type_name
        self.output_origin = self.img_origin_during_run

        # FIXME put out of the class ?
        if vol_type == "Image":
            if self.user_info.img_coord_system:
                vol = self.volume_engine.g4_volumes[vol_name]
                # Translate the output dose map so that its center correspond to the image center.
                # The origin is thus the center of the first voxel.
                img_info = gate.get_info_from_image(vol.image)
                dose_info = gate.get_info_from_image(self.py_numerator_image)
                self.output_origin = gate.get_origin_wrt_images_g4_position(
                    img_info, dose_info, self.user_info.translation
                )
        else:
            if self.user_info.img_coord_system:
                gate.warning(
                    f'LETActor "{self.user_info.name}" has '
                    f"the flag img_coord_system set to True, "
                    f"but it is not attached to an Image "
                    f'volume ("{vol_name}", of type "{vol_type}"). '
                    f"So the flag is ignored."
                )
        # user can set the output origin
        if self.user_info.output_origin is not None:
            if self.user_info.img_coord_system:
                gate.warning(
                    f'LETActor "{self.user_info.name}" has '
                    f"the flag img_coord_system set to True, "
                    f"but output_origin is set, so img_coord_system ignored."
                )
            self.output_origin = self.user_info.output_origin

    def EndSimulationAction(self):
        g4.GateLETActor.EndSimulationAction(self)

        # Get the itk image from the cpp side
        # Currently a copy. Maybe latter as_pyarray ?
        self.py_numerator_image = gate.get_cpp_image(self.cpp_numerator_image)
        self.py_denominator_image = gate.get_cpp_image(self.cpp_denominator_image)

        # set the property of the output image:
        # in the coordinate system of the attached volume
        # FIXME no direction for the moment ?
        self.py_numerator_image.SetOrigin(self.output_origin)
        # self.py_denominator_image.SetOrigin(self.output_origin)

        # write the image at the end of the run
        # FIXME : maybe different for several runs
        if self.user_info.output:
            suffix = ""
            if self.user_info.dose_average:
                suffix += "_letd"
            elif self.user_info.track_average:
                suffix += "_lett"
            if self.user_info.let_to_other_material or self.user_info.let_to_water:
                suffix += f"_convto_{self.user_info.other_material}"

            fPath = str(self.user_info.output).replace(".mhd", f"{suffix}.mhd")
            self.user_info.output = fPath
            # self.output = fPath
            self.py_LETd_image = gate.divide_itk_images(
                img1_numerator=self.py_numerator_image,
                img2_denominator=self.py_denominator_image,
                filterVal=0,
                replaceFilteredVal=0,
            )
            itk.imwrite(self.py_LETd_image, gate.check_filename_type(fPath))

            # for parrallel computation we need to provide both outputs
            if self.user_info.separate_output:
                fPath = fPath.replace(".mhd", "_numerator.mhd")
                itk.imwrite(self.py_numerator_image, gate.check_filename_type(fPath))
                fPath = fPath.replace("_numerator", "_denominator")
                itk.imwrite(self.py_denominator_image, gate.check_filename_type(fPath))

        # debug
        """itk.imwrite(self.py_square_image, "square.mhd")
        itk.imwrite(self.py_temp_image, "temp.mhd")
        itk.imwrite(self.py_last_id_image, "lastid.mhd")
        itk.imwrite(self.uncertainty_image, "uncer.mhd")"""



import opengate as gate
import opengate_core as g4
from box import Box
from scipy.spatial.transform import Rotation
import numpy as np


class RepeatParametrisedVolume(gate.VolumeBase):
    """
    Allow to repeat a volume with translations
    """

    type_name = "RepeatParametrised"

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        user_info.material = "G4_AIR"
        user_info.repeated_volume_name = None
        user_info.linear_repeat = None
        user_info.translation = None
        user_info.rotation = Rotation.identity().as_matrix()
        user_info.offset = [0, 0, 0]
        user_info.start = None
        user_info.offset_nb = 1

    def __init__(self, user_info):
        super().__init__(user_info)

    def __del__(self):
        pass

    def construct_solid(self):
        # no solid to build
        pass

    def construct_logical_volume(self):
        # check
        if self.user_info.repeated_volume_name is None:
            gate.fatal(
                f'Repeater "{self.user_info.name}": the option repeated_volume_name must be set'
            )
        if self.user_info.linear_repeat is None:
            gate.fatal(
                f'Repeater "{self.user_info.name}": the option linear_repeat must be set'
            )
        # the repeated volume *must* have been build before
        v = self.volume_engine.get_volume(self.user_info.repeated_volume_name, False)
        # check phys vol
        if v.user_info.build_physical_volume:
            gate.fatal(
                f"Error ! the volume {v.user_info.name} already have a physical volume. "
                f'Set "build_physical_volume" to False'
            )
        if v.g4_physical_volume:
            gate.fatal(
                f"Error ! the volume {v.user_info.name} already have a physical volume. "
                f'Set "build_physical_volume" to False'
            )
        # set log vol
        self.g4_logical_volume = v.g4_logical_volume

    def construct_physical_volume(self):
        # find the mother's logical volume
        st = g4.G4LogicalVolumeStore.GetInstance()
        mother_logical = st.GetVolume(self.user_info.mother, False)
        if not mother_logical:
            gate.fatal(f"The mother of {self.user_info.name} cannot be the world.")

        # create parameterised
        p = Box()
        p.linear_repeat = self.user_info.linear_repeat
        p.start = self.user_info.start
        p.translation = self.user_info.translation
        p.rotation = self.user_info.rotation
        if (  # FIXME put in helper functions
            p.rotation is None
            or not isinstance(p.rotation, (np.matrix, np.ndarray))
            or p.rotation.shape != (3, 3)
        ):
            p.rotation = Rotation.identity().as_matrix()
        p.offset = self.user_info.offset
        p.offset_nb = self.user_info.offset_nb
        self.param = g4.GateRepeatParameterisation()
        self.param.SetUserInfo(p)

        # number of copies
        n = p.linear_repeat[0] * p.linear_repeat[1] * p.linear_repeat[2] * p.offset_nb

        # (only daughter)
        # g4.EAxis.kUndefined => faster
        self.g4_physical_volume = g4.G4PVParameterised(
            self.user_info.name,
            self.g4_logical_volume,
            mother_logical,
            g4.EAxis.kUndefined,
            n,
            self.param,
            False,
        )  # very long if True

        self.g4_physical_volumes.append(self.g4_physical_volume)



import opengate_core as g4
import opengate as gate


class ExceptionHandler(g4.GateExceptionHandler):
    """
    Geant4 exception handler. Inherit form GateExceptionHandler
    The function 'Notify' will be called by G4 when an exception occurs.
    Should be created after initialization.
    Will be automatically stoed in the G4 stateManager
    """

    def Notify(self, originOfException, exceptionCode, severity, description):
        s = f"G4Exception origin: {originOfException}\n"
        s += f"G4Exception code: {exceptionCode}\n"
        s += f"G4Exception severity: {severity}\n"
        s += f"G4Exception: {description}"
        if severity == g4.FatalException or severity == g4.FatalErrorInArgument:
            gate.fatal(s)
        gate.warning(s)
        return False



from .DigitizerHitsCollectionActor import *


class ARFTrainingDatasetActor(g4.GateARFTrainingDatasetActor, gate.ActorBase):
    """
    The ARFTrainingDatasetActor build a root file with energy, angles, positions and energy windows
    of a spect detector. To be used by garf_train to train a ARF neural network.

    Note: Must inherit from ActorBase not from HitsCollectionActor, even if the
    cpp part inherit from HitsCollectionActor
    """

    type_name = "ARFTrainingDatasetActor"

    @staticmethod
    def set_default_user_info(user_info):
        DigitizerHitsCollectionActor.set_default_user_info(user_info)
        user_info.attributes = []
        user_info.output = "arf_training.root"
        user_info.debug = False
        user_info.energy_windows_actor = None
        user_info.russian_roulette = 1

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GateARFTrainingDatasetActor.__init__(self, user_info.__dict__)

    def __del__(self):
        pass

    def __str__(self):
        s = f"ARFTrainingDatasetActor {self.user_info.name}"
        return s



from box import Box
import opengate as gate
import opengate_core as g4


class VolumeEngine(g4.G4VUserDetectorConstruction, gate.EngineBase):
    """
    FIXME
    """

    def __init__(self, simulation):
        g4.G4VUserDetectorConstruction.__init__(self)
        gate.EngineBase.__init__(self)

        # keep input data
        self.simulation = simulation
        self.volume_manager = simulation.volume_manager
        self.is_constructed = False
        self.actor_engine = None

        # tree of volumes
        self.volumes_tree = None
        self.g4_volumes = {}

    def __del__(self):
        if self.verbose_destructor:
            print("del VolumeEngine")
        pass

    def Construct(self):
        """
        G4 overloaded.
        Override the Construct method from G4VUserDetectorConstruction
        """

        # build the tree of volumes
        self.simulation.check_geometry()
        self.volumes_tree = gate.build_tree(self.simulation)

        # build all G4 volume objects
        self.build_g4_volumes()

        # return the world physical volume
        self.is_constructed = True
        return self.g4_volumes[gate.__world_name__].g4_physical_volume

    def check_overlaps(self, verbose):
        for v in self.g4_volumes.values():
            for w in v.g4_physical_volumes:
                try:
                    b = w.CheckOverlaps(1000, 0, verbose, 1)
                    if b:
                        gate.fatal(
                            f'Some volumes overlap the volume "{v}". \n'
                            f"Consider using G4 verbose to know which ones. \n"
                            f"Aborting."
                        )
                except:
                    pass
                    # gate.warning(f'do not check physical volume {w}')

    def find_or_build_material(self, material):
        mat = self.volume_manager.material_database.FindOrBuildMaterial(material)
        return mat

    def build_g4_volumes(self):
        uiv = self.volume_manager.user_info_volumes
        for vu in uiv.values():
            # create the volume
            vol = gate.new_element(vu, self.simulation)
            # construct the G4 Volume
            vol.construct(self)
            if len(vol.g4_physical_volumes) == 0:
                vol.g4_physical_volumes.append(vol.g4_physical_volume)
            # keep the volume
            self.g4_volumes[vu.name] = vol

    def set_actor_engine(self, actor_engine):
        self.actor_engine = actor_engine

    def ConstructSDandField(self):
        """
        G4 overloaded
        """
        # This function is called in MT mode
        tree = self.volumes_tree
        self.actor_engine.register_sensitive_detectors(tree)

    def get_volume(self, name, check_initialization=True):
        if check_initialization and not self.is_constructed:
            gate.fatal(f"Cannot get_volume before initialization")
        if name not in self.g4_volumes:
            gate.fatal(
                f"The volume {name} is not in the current "
                f"list of volumes: {self.g4_volumes}"
            )
        return self.g4_volumes[name]

    def get_database_material_names(self, db=None):
        return self.volume_manager.material_database.get_database_material_names(db)

    def dump_build_materials(self, level=0):
        table = g4.G4Material.GetMaterialTable
        if level == 0:
            names = [m.GetName() for m in table]
            return names
        return table



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

paths = gate.get_default_test_paths(__file__, "gate_test004_simulation_stats_actor")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.running_verbose_level = 0
ui.visu = False
ui.number_of_threads = 5
# special debug mode : force MT even with one single thread
ui.force_multithread_mode = True
ui.random_engine = "MixMaxRng"
ui.random_seed = "auto"

"""
    Warning: we can only see the speed up of the MT mode for large nb of particles (>2e6)
"""

# set the world size like in the Gate macro
m = gate.g4_units("m")
world = sim.world
world.size = [3 * m, 3 * m, 3 * m]
world.material = "G4_AIR"

# add a simple waterbox volume
waterbox = sim.add_volume("Box", "Waterbox")
cm = gate.g4_units("cm")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
waterbox.material = "G4_WATER"

# physic list
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMV"
cuts = p.production_cuts
um = gate.g4_units("um")
cuts.world.gamma = 700 * um
cuts.world.electron = 700 * um
cuts.world.positron = 700 * um
cuts.world.proton = 700 * um

# default source for tests
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
source = sim.add_source("GenericSource", "Default")
source.particle = "gamma"
source.energy.mono = 80 * keV
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.n = 200000 / ui.number_of_threads

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# start simulation
# sim.apply_g4_command("/run/verbose 0")
# sim.apply_g4_command("/run/eventModulo 5000 1")
output = sim.start()

# get results
stats = output.get_actor("Stats")
print(stats)
print("track type", stats.counts.track_types)

# gate_test4_simulation_stats_actor
# Gate mac/main.mac
stats_ref = gate.read_stat_file(paths.gate_output / "stat.txt")
stats_ref.counts.run_count = sim.user_info.number_of_threads
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.01)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate.contrib.pet_siemens_biograph as pet_biograph
import opengate as gate
import opengate.contrib.phantom_necr as phantom_necr
from test037_pet_hits_singles_helpers import (
    default_root_hits_branches,
    default_root_singles_branches,
)
import uproot
import numpy as np

paths = gate.get_default_test_paths(__file__, "gate_test049_pet_blur")


def create_simulation(sim, threads=1, singles_name="Singles"):
    # main options
    sim.user_info.visu = False
    sim.user_info.number_of_threads = threads
    sim.user_info.random_seed = 123456789

    # units
    m = gate.g4_units("m")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    MBq = Bq * 1e6
    sec = gate.g4_units("second")

    #  change world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]
    world.material = "G4_AIR"

    # add a PET Biograph
    pet = pet_biograph.add_pet(sim, "pet")
    singles = pet_biograph.add_digitizer(
        sim, pet.name, paths.output / f"test049_pet.root", singles_name=singles_name
    )

    # add NECR phantom
    phantom = phantom_necr.add_necr_phantom(sim, "phantom")

    # physics
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    sim.set_cut("world", "all", 1 * m)
    sim.set_cut(phantom.name, "all", 10 * mm)
    sim.set_cut(f"{pet.name}_crystal", "all", 0.1 * mm)

    # default source for tests
    source = phantom_necr.add_necr_source(sim, phantom)
    total_yield = gate.get_rad_yield("F18")
    print("Yield for F18 (nb of e+ per decay) : ", total_yield)
    source.activity = 3000 * Bq * total_yield
    source.activity = 1787.914158 * MBq * total_yield / sim.user_info.number_of_threads
    # source.n = 50000
    source.half_life = 6586.26 * sec
    source.energy.type = "F18_analytic"  # WARNING not ok, but similar to previous Gate
    # source.energy.type = "F18"  # this is the correct F18 e+ source

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    # timing
    sec = gate.g4_units("second")
    sim.run_timing_intervals = [[0, 0.00005 * sec]]
    # sim.run_timing_intervals = [[0, 0.00005 * sec]]


def check_root_hits(paths, nb, ref_hits_output, hits_output, png_output="auto"):
    if png_output == "auto":
        png_output = f"test037_test{nb}_hits.png"
    # check phsp (new version)
    print()
    gate.warning(f"Check root (hits)")
    k1, k2 = default_root_hits_branches()
    p1 = gate.root_compare_param_tree(ref_hits_output, "Hits", k1)
    # in the legacy gate, some edep=0 are still saved in the root file,
    # so we don't count that ones in the histogram comparison
    p1.mins[k1.index("edep")] = 0
    p2 = gate.root_compare_param_tree(hits_output, "Hits", k2)
    # p2.scaling[p2.the_keys.index("GlobalTime")] = 1e-9  # time in ns
    p1.scaling[p1.the_keys.index("time")] = 1e9  # time in ns
    p = gate.root_compare_param(p1.the_keys, paths.output / png_output)
    p.hits_tol = 6  # % tolerance (including the edep zeros)
    p.tols[k1.index("posX")] = 12
    p.tols[k1.index("posY")] = 13
    p.tols[k1.index("posZ")] = 2.1
    p.tols[k1.index("edep")] = 0.003
    p.tols[k1.index("time")] = 480
    is_ok = gate.root_compare4(p1, p2, p)

    return is_ok


def check_root_singles(
    paths, v, ref_singles_output, singles_output, sname="Singles", png_output="auto"
):
    if png_output == "auto":
        png_output = f"test037_test{v}_singles.png"
    # check phsp (singles)
    print()
    gate.warning(f"Check root (singles)")
    k1, k2 = default_root_singles_branches()
    p1 = gate.root_compare_param_tree(ref_singles_output, "Singles", k1)
    # in the legacy gate, some edep=0 are still saved in the root file,
    # so we don't count that ones in the histogram comparison
    p1.mins[k1.index("energy")] = 0
    p2 = gate.root_compare_param_tree(singles_output, sname, k2)
    # p2.scaling[p2.the_keys.index("GlobalTime")] = 1e-9  # time in ns
    p1.scaling[p1.the_keys.index("time")] = 1e9  # time in ns
    p = gate.root_compare_param(p1.the_keys, paths.output / png_output)
    p.hits_tol = 5  # % tolerance (including the edep zeros)
    p.tols[k1.index("globalPosX")] = 8
    p.tols[k1.index("globalPosY")] = 5
    p.tols[k1.index("globalPosZ")] = 1.1
    p.tols[k1.index("energy")] = 0.0045
    p.tols[k1.index("time")] = 350

    is_ok = gate.root_compare4(p1, p2, p)

    return is_ok


def check_timing(
    ref_root_file,
    root_file,
):
    times_ref = uproot.open(ref_root_file)["Hits"].arrays(library="numpy")["time"] * 1e9
    times = uproot.open(root_file)["Hits"].arrays(library="numpy")["GlobalTime"]

    def rel_d(a, b, norm, tol):
        r = np.fabs(a - b) / norm * 100
        s = f"{a:.2f} {b:.2f} {r:.2f}% "
        is_ok = r < tol
        return s, is_ok

    def compare_stat(ref, val, tol):
        m = np.mean(ref)
        s1, is_ok1 = rel_d(np.min(ref), np.min(val), m, tol)
        s2, is_ok2 = rel_d(m, np.mean(val), m, tol)
        s3, is_ok3 = rel_d(np.max(ref), np.max(val), m, tol)
        s = f"{s1}    {s2}     {s3}"
        return s, is_ok1 and is_ok2 and is_ok3

    tol = 1
    s, b = compare_stat(times_ref, times, tol)
    print()
    gate.print_test(b, f"Hits timing ref : {s}")
    is_ok = b

    times_ref = (
        uproot.open(ref_root_file)["Singles"].arrays(library="numpy")["time"] * 1e9
    )
    times = uproot.open(root_file)["Singles"].arrays(library="numpy")["GlobalTime"]

    print()
    s, b = compare_stat(times_ref, times, tol)
    gate.print_test(b, f"Singles timing ref : {s}")
    is_ok = is_ok and b

    print()
    min_ref = np.min(times_ref)
    min_v = np.min(times)
    tol = -10
    b = min_v < tol
    gate.print_test(b, f"Compare time min values : {min_ref} vs {min_v} wrt {tol}")

    return is_ok and b



import opengate as gate


class UserElement:
    """
    Common class for all types of elements (volume, source or actor)
    Manager a dict (Box) for user parameters: user_info
    Check that all the required keys are provided
    """

    def __init__(self, user_info):
        # set the user info (a kind of dict)
        self.user_info = user_info
        # check everything is there (except for solid building)
        self.check_user_info()
        # check type_name
        if self.user_info.type_name != self.type_name:
            gate.fatal(
                f"Error, the type_name inside the user_info is different "
                f"from the type_name of the class: {self.user_info} in the "
                f"class {self.__name__} {self.type_name}"
            )
        # by default the name is a unique id (uuid)
        if not self.user_info.name:
            gate.fatal(
                f"Error a {self.user_info.volume_type} must have "
                f"a valid name, while it is {self.user_info.name}"
            )

    @staticmethod
    def set_default_user_info(user_info):
        # Should be overwritten by subclass
        pass

    def __del__(self):
        pass

    def __str__(self):
        s = f"Element: {self.user_info}"
        return s

    def set_simulation(self, simulation):
        self.simulation = simulation

    def check_user_info(self):
        # get a fake ui to compare
        ref_ui = gate.UserInfo(self.user_info.element_type, self.user_info.type_name)
        # if this is a solid, we do not check some keys (mother, translation etc)
        if "i_am_a_solid" in self.user_info.__dict__:
            gate.VolumeManager._pop_keys_unused_by_solid(ref_ui)
        for val in ref_ui.__dict__:
            if val not in self.user_info.__dict__:
                gate.fatal(f'Cannot find "{val}" in {self.user_info}')
        for val in self.user_info.__dict__:
            # special case for solid, and boolean
            if val == "i_am_a_solid" or val == "solid":
                continue
            if val == "nodes" or val == "add_node":
                continue
            if val not in ref_ui.__dict__.keys():
                gate.warning(f'Unused param "{val}" in {self.user_info}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test014_engine_helpers import *

sim = gate.Simulation()
define_simulation(sim)

# go with a new process
se = gate.SimulationEngine(sim, start_new_process=True)
output = se.start()

# get output
is_ok = test_output(output)

# go without a new process
se = gate.SimulationEngine(sim, start_new_process=False)
output = se.start()

# get output
is_ok = test_output(output) and is_ok

gate.test_ok(is_ok)



import opengate as gate
from scipy.spatial.transform import Rotation

# colors (similar to the ones of Gate)
red = [1, 0, 0, 1]
blue = [0, 0, 1, 1]
green = [0, 1, 0, 1]
yellow = [0.9, 0.9, 0.3, 1]
gray = [0.5, 0.5, 0.5, 1]
white = [1, 1, 1, 0.8]


def add_necr_phantom(sim, name="necr"):
    """
    Simple NECR phantom with a cylinder and linear source
    (geometry part only, the source is defined in the other function)
    """

    # unit
    mm = gate.g4_units("mm")
    cm = gate.g4_units("cm")

    # ring volume
    phantom = sim.add_volume("Tubs", name)
    phantom.mother = "world"
    phantom.rmax = 103 * mm
    phantom.rmin = 0 * mm
    phantom.dz = 71 * cm / 2.0
    phantom.material = "G4_AIR"
    phantom.color = gray

    # polyethylene cylinder scat
    cylinderScat = sim.add_volume("Tubs", f"{name}_cylinderScat")
    cylinderScat.mother = phantom.name
    cylinderScat.translation = [0, 0, 0]
    cylinderScat.rmax = 102 * mm
    cylinderScat.rmin = 0 * mm
    cylinderScat.dz = 70 * cm / 2.0
    cylinderScat.material = "G4_POLYETHYLENE"
    cylinderScat.color = gray

    # line source interior
    linear_source_in = sim.add_volume("Tubs", f"{name}_linear_source_in")
    linear_source_in.mother = cylinderScat.name
    linear_source_in.translation = [0, -4.5 * cm, 0]
    linear_source_in.rmax = 1.6 * mm
    linear_source_in.rmin = 0 * mm
    linear_source_in.dz = 70 * cm / 2.0
    linear_source_in.material = "G4_WATER"
    linear_source_in.color = red

    # line source exterior
    linear_source_out = sim.add_volume("Tubs", f"{name}_linear_source_out")
    linear_source_out.mother = cylinderScat.name
    linear_source_out.translation = [0, -4.5 * cm, 0]
    linear_source_out.rmax = 2.5 * mm
    linear_source_out.rmin = 1.6 * mm
    linear_source_out.dz = 70 * cm / 2.0
    linear_source_out.material = "G4_POLYETHYLENE"
    linear_source_out.color = red

    return phantom


def add_necr_source(sim, necr_phantom):
    """
    The source is attached to the linear_source_in volume,
    it means its coordinate system is the same
    (not the shape).
    """

    v = sim.get_volume_user_info(f"{necr_phantom.name}_linear_source_in")

    src = sim.add_source("GenericSource", f"{necr_phantom.name}_source")
    src.mother = v.name
    src.particle = "e+"
    src.energy.type = "F18"
    src.position.type = "cylinder"
    src.position.radius = v.rmax
    src.position.dz = v.dz
    src.direction.type = "iso"

    return src


def add_necr_source_debug(sim, name):
    """
    The source is described according to the world
    """

    mm = gate.g4_units("mm")
    cm = gate.g4_units("cm")
    MeV = gate.g4_units("MeV")
    src = sim.add_source("GenericSource", name)

    src.particle = "gamma"
    src.energy.type = "mono"
    src.energy.mono = 0.511 * MeV
    src.position.type = "sphere"
    src.position.radius = 1.6 * mm

    """
    src.particle = "e+"
    src.energy.type = "F18_analytic"
    src.position.type = "cylinder"
    src.position.radius = 1.6 * mm
    src.position.dz = 35 * cm
    src.position.translation = [0, -4.5 * cm, 0]
    """
    src.direction.type = "iso"

    return src



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test029_volume_time_rotation_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test029_volume_time_rotation")

# create the main simulation object
sim = gate.Simulation()

# create sim without AA
create_simulation(sim, False)

# initialize & start
output = sim.start(True)

"""
# use to create the (fake) reference for test029_volume_time_rotation_2.py
import itk
scaling = 0.90
img = itk.imread(str(paths.output_ref / "proj029.mhd"))
arr = itk.array_view_from_image(img)
arr *= scaling
itk.imwrite(img, str(paths.output_ref / "proj029_scaled.mhd"))
"""

# -------------------------
gate.warning("Compare stats")
stats = gate.read_stat_file(paths.output / "stats029.txt")
print(stats)
stats_ref = gate.read_stat_file(paths.output_ref / "stats029.txt")
print(
    f"Number of steps was {stats.counts.step_count}, forced to the same value (because of angle acceptance). "
)
stats.counts.step_count = stats_ref.counts.step_count  # force to id
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.01)
print(is_ok)

gate.warning("Compare images")
# read image and force change the offset to be similar to old Gate
is_ok = (
    gate.assert_images(
        paths.output_ref / "proj029.mhd",
        paths.output / "proj029.mhd",
        stats,
        tolerance=47,
        ignore_value=0,
        axis="x",
        sum_tolerance=0.5,
    )
    and is_ok
)
print(is_ok)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test029_volume_time_rotation_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test029_volume_time_rotation")

# create the main simulation object
sim = gate.Simulation()

# create sim without AA
create_simulation(sim, True)

# initialize & start
output = sim.start()

"""
WARNING when "angle_acceptance_volume" is enabled, it is a bit faster (+50%) but the result is not
exactly the same as without. This is because, even if the initial particle is not in the direction of
the spect system, it can scatter and still reach the detector.

We don't have the collimator here (to
faster the simulation), this is why the difference is not negligible.

The (fake) reference is computed by scaling the ref by 90% (test029_volume_time_rotation_1.py)
"""

gate.warning("Compare stats")
stats = gate.read_stat_file(paths.output / "stats029.txt")
print(stats)
stats_ref = gate.read_stat_file(paths.output_ref / "stats029.txt")
print(
    f"Number of steps was {stats.counts.step_count}, forced to the same value (because of angle acceptance). "
)
stats.counts.step_count = stats_ref.counts.step_count  # force to id
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.01)
print(is_ok)

gate.warning("Compare images")
# read image and force change the offset to be similar to old Gate
is_ok = (
    gate.assert_images(
        paths.output_ref / "proj029_scaled.mhd",
        paths.output / "proj029.mhd",
        stats,
        tolerance=50,
        ignore_value=0,
        axis="x",
        sum_tolerance=1,
    )
    and is_ok
)
print(is_ok)

gate.test_ok(is_ok)



from .GenericSource import *
from .GANSourceDefaultGenerator import GANSourceDefaultGenerator
import time
import scipy


class GANSourceDefaultPairsGenerator(GANSourceDefaultGenerator):
    """
    Like GANSourceDefaultGenerator but for pairs of particle (PET)
    """

    def __init__(self, user_info):
        super().__init__(user_info)
        self.is_paired = True

    def __getstate__(self):
        self.lock = None
        self.gaga = None
        self.gan_info = None
        return self.__dict__

    def check_parameters(self, g):
        # position
        if g.position_is_set_by_GAN and len(self.user_info.position_keys) != 6:
            dim = len(self.user_info.position_keys)
            self.fatal(f"you must provide 6 values for position, while it was {dim}")
        # direction
        if g.direction_is_set_by_GAN and len(self.user_info.direction_keys) != 6:
            dim = len(self.user_info.direction_keys)
            self.fatal(f"you must provide 6 values for direction, while it was {dim}")

    def get_energy_index(self, g, the_keys, n):
        # get energy index from GAN
        if not g.energy_is_set_by_GAN:
            return
        ek = self.user_info.energy_key
        dim = len(ek)
        if dim != 2:
            self.fatal(f"you must provide 2 values for energy, while it was {dim}")
        g.energy_gan_index = [the_keys.index(ek[0]), the_keys.index(ek[1])]

    def get_time_index(self, g, the_keys, n):
        # get time index from GAN
        if not g.time_is_set_by_GAN:
            return
        ek = self.user_info.time_key
        dim = len(ek)
        if dim != 2:
            self.fatal(f"you must provide 2 values for time, while it was {dim}")
        g.time_gan_index = [the_keys.index(ek[0]), the_keys.index(ek[1])]

    def get_weight_index(self, g, the_keys, n):
        # get weight index from GAN
        if not g.weight_is_set_by_GAN:
            return
        ek = self.user_info.weight_key
        dim = len(ek)
        if dim != 2:
            self.fatal(f"you must provide 2 values for weight, while it was {dim}")
        g.weight_gan_index = [the_keys.index(ek[0]), the_keys.index(ek[1])]

    def generator(self, source):
        """
        Main function that will be called from the cpp side every time a batch
        of particles should be created.
        Once created here, the particles are copied to cpp.
        (Yes maybe the copy could be avoided, but I did not manage to do it)
        """
        # get the info
        g = self.gan_info
        n = self.user_info.batch_size
        start = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start = time.time()
            print(f"Generate {n} particles from GAN ", end="")

        # generate samples (this is the most time-consuming part)
        fake = self.gaga.generate_samples2(
            g.params,
            g.G,
            g.D,
            n=n,
            batch_size=n,
            normalize=False,
            to_numpy=True,
            silence=True,
        )

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # move particle backward ?
        self.move_backward(g, fake)

        # copy to cpp
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(f"in {end - start:0.1f} sec (GPU={g.params.current_gpu})")

    def copy_generated_particle_to_g4(self, source, g, fake):
        # position
        if g.position_is_set_by_GAN:
            pos = []
            dim = len(g.position_gan_index)
            for i in range(dim):
                if g.position_use_index[i]:
                    pos.append(fake[:, g.position_gan_index[i]])
                else:
                    pos.append(g.position_gan_index[i])
            # copy to c++
            source.fPositionX = pos[0]
            source.fPositionY = pos[1]
            source.fPositionZ = pos[2]
            source.fPositionX2 = pos[3]
            source.fPositionY2 = pos[4]
            source.fPositionZ2 = pos[5]

        # direction
        if g.direction_is_set_by_GAN:
            dir = []
            dim = len(g.direction_gan_index)
            for i in range(dim):
                if g.direction_use_index[i]:
                    dir.append(fake[:, g.direction_gan_index[i]])
                else:
                    dir.append(g.direction_gan_index[i])
            # copy to c++
            source.fDirectionX = dir[0]
            source.fDirectionY = dir[1]
            source.fDirectionZ = dir[2]
            source.fDirectionX2 = dir[3]
            source.fDirectionY2 = dir[4]
            source.fDirectionZ2 = dir[5]

        # energy
        if g.energy_is_set_by_GAN:
            # copy to c++
            source.fEnergy = fake[:, g.energy_gan_index[0]]
            source.fEnergy2 = fake[:, g.energy_gan_index[1]]

        # time
        if g.time_is_set_by_GAN:
            # copy to c++
            source.fTime = fake[:, g.time_gan_index[0]]
            source.fTime2 = fake[:, g.time_gan_index[1]]

        # weight
        if g.weight_is_set_by_GAN:
            # copy to c++
            source.fWeight = fake[:, g.weight_gan_index[0]]
            source.fWeight2 = fake[:, g.weight_gan_index[1]]

    def move_backward(self, g, fake):
        # move particle backward ?
        back = self.user_info.backward_distance
        if not back:
            return

        if not g.time_is_set_by_GAN and not self.user_info.backward_force:
            gate.fatal(
                f"If backward is enabled the time is not managed by GAN,"
                f" time is wrong. IT can be forced, however, with the option 'backward_force'"
            )

        # move particle position
        position = fake[:, g.position_gan_index[0] : g.position_gan_index[0] + 3]
        direction = fake[:, g.direction_gan_index[0] : g.direction_gan_index[0] + 3]
        fake[:, g.position_gan_index[0] : g.position_gan_index[0] + 3] = (
            position - back * direction
        )

        # move second particle position
        position = fake[:, g.position_gan_index[3] : g.position_gan_index[3] + 3]
        direction = fake[:, g.direction_gan_index[3] : g.direction_gan_index[3] + 3]
        fake[:, g.position_gan_index[3] : g.position_gan_index[3] + 3] = (
            position - back * direction
        )

        # modify the time because we move the particle backward
        if g.time_is_set_by_GAN:
            c = scipy.constants.speed_of_light * 1000 / 1e9  # in mm/ns
            xt = back / c
            fake[:, g.time_gan_index[0]] -= xt
            fake[:, g.time_gan_index[1]] -= xt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4
from scipy.spatial.transform import Rotation
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

# global log level
# create the simulation
sim = gate.Simulation()
print(f"Volumes types: {sim.dump_volume_types()}")

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.check_volumes_overlap = True

# add a material database
sim.add_material_database(pathFile / ".." / "data" / "GateMaterials.db")

#  change world size
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# create a union of several volumes

# first create the solids
b = sim.new_solid("Box", "box")
b.size = [10 * cm, 10 * cm, 10 * cm]
s = sim.new_solid("Sphere", "sphere")
s.rmax = 5 * cm
t = sim.new_solid("Tubs", "t")
t.rmin = 0
t.rmax = 2 * cm
t.dz = 15 * cm

# bool operations
a = gate.solid_union(b, s, [0, 1 * cm, 5 * cm])
a = gate.solid_subtraction(a, t, [0, 1 * cm, 5 * cm])
a = gate.solid_union(a, b, [0, -1 * cm, -5 * cm])  # strange but ok
b = gate.solid_intersection(t, s, [3 * cm, 0, 0])
a = gate.solid_union(a, b, [0, -7 * cm, -5 * cm])

# then add them to a Union, with translation/rotation
rot = Rotation.from_euler("x", 33, degrees=True).as_matrix()
u = sim.add_volume_from_solid(a, "my_stuff")
u.translation = [5 * cm, 5 * cm, 5 * cm]
u.rotation = rot
u.mother = "world"
u.material = "G4_WATER"
u.color = [0, 1, 0, 1]

# create a volume from a solid (not really useful)
u = sim.add_volume_from_solid(s, "test_sph")
u.translation = [-5 * cm, -5 * cm, 1 - 5 * cm]
u.mother = "world"
u.material = "G4_WATER"
u.color = [0, 1, 1, 1]

# default source for tests
source = sim.add_source("GenericSource", "Default")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
source.particle = "proton"
source.energy.mono = 240 * MeV
source.position.radius = 1 * cm
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 5 * Bq

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")


# function to run after init
def after_init(se):
    print("Checking solid ...")
    ve = se.volume_engine
    v = ve.get_volume("my_stuff")
    v = v.g4_logical_volume
    is_ok = v.GetName() == "my_stuff"
    gate.print_test(is_ok, f"Get volume {v.GetName()}")
    solid = v.GetSolid()
    pMin = g4.G4ThreeVector()
    pMax = g4.G4ThreeVector()
    solid.BoundingLimits(pMin, pMax)
    is_ok = list(pMin) == list([-50, -90, -100]) and is_ok
    gate.print_test(is_ok, f"pMin {pMin}")
    is_ok = list(pMax) == list([50, 60, 100]) and is_ok
    gate.print_test(is_ok, f"pMax {pMax}")
    if not is_ok:
        gate.test_ok(is_ok)


# create G4 objects
print(sim)

# start simulation
se = gate.SimulationEngine(sim)
se.user_fct_after_init = after_init
output = se.start()

# print results at the end
stats = output.get_actor("Stats")
print(stats)

gate.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "gate_test009_voxels")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.number_of_threads = 1
print(ui)

# add a material database
sim.add_material_database(paths.data / "GateMaterials.db")

#  change world size
m = gate.g4_units("m")
mm = gate.g4_units("mm")
um = gate.g4_units("um")
keV = gate.g4_units("keV")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# add a simple fake volume to test hierarchy
# translation and rotation like in the Gate macro
fake = sim.add_volume("Box", "fake")
cm = gate.g4_units("cm")
fake.size = [40 * cm, 40 * cm, 40 * cm]
fake.material = "G4_WATER"
fake.color = [1, 0, 1, 1]
fake.rotation = Rotation.from_euler("x", 2, degrees=True).as_matrix()

# image
patient = sim.add_volume("Image", "patient")
patient.image = paths.data / "patient-4mm.mhd"
patient.mother = "fake"
patient.material = "G4_AIR"  # default material
vm = gate.read_voxel_materials(paths.gate_data / "patient-HU2mat-v1.txt")
vm[0][0] = -10000
patient.voxel_materials = vm
patient.dump_label_image = paths.output / "test020_labels.mhd"

# activity
activity = 100 * kBq

# source 1
source = sim.add_source("GenericSource", "source1")
source.energy.mono = 150 * keV
source.particle = "gamma"
source.position.type = "sphere"
source.position.radius = 10 * mm
source.position.translation = [0, 0, -15 * cm]
source.activity = activity / ui.number_of_threads
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]

# large cuts, no e- needed
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMV"
c = p.production_cuts
c.world.gamma = 700 * um
c.world.positron = 1 * mm
c.world.electron = 1 * m
c.world.proton = 1 * m

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test20-edep.mhd"
dose.mother = "patient"
dose.size = [100, 100, 100]
dose.spacing = [2 * mm, 2 * mm, 2 * mm]
dose.img_coord_system = True  # default is True
dose.translation = [0 * mm, 0 * mm, 1 * mm]

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

# verbose
sim.apply_g4_command("/tracking/verbose 0")

# start simulations
output = sim.start()

# print results at the end
stat = output.get_actor("Stats")
print(stat)
d = output.get_actor("dose")
print(d)

# tests
stats_ref = gate.read_stat_file(paths.gate / "output" / "stat_profiling.txt")
stats_ref.counts.run_count = ui.number_of_threads
is_ok = gate.assert_stats(stat, stats_ref, 0.1)
is_ok = is_ok and gate.assert_images(
    paths.gate / "output" / "output_profiling-Edep.mhd",
    paths.output / "test20-edep.mhd",
    stat,
    tolerance=79,
)
gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
import os

paths = gate.get_default_test_paths(__file__, "gate_test044_pbs")
output_path = paths.output / "output_test044_weight"

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_seed = 123654789
ui.random_engine = "MersenneTwister"

# units
km = gate.g4_units("km")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
um = gate.g4_units("um")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
nm = gate.g4_units("nm")
deg = gate.g4_units("deg")
mrad = gate.g4_units("mrad")

# add a material database
sim.add_material_database(paths.gate_data / "HFMaterials2014.db")

#  change world size
world = sim.world
world.size = [600 * cm, 500 * cm, 500 * cm]

## FIRST SOURCE + DETECTOR ##
# waterbox
# translation and rotation like in the Gate macro
waterbox1 = sim.add_volume("Box", "waterbox1")
waterbox1.size = [10 * cm, 10 * cm, 105 * cm]
waterbox1.translation = [0 * cm, 0 * cm, 52.5 * cm]
waterbox1.material = "Vacuum"
waterbox1.color = [0, 0, 1, 1]

# phantoms
m = Rotation.identity().as_matrix()

phantom = sim.add_volume("Box", "phantom_a_1")
phantom.mother = "waterbox1"
phantom.size = [100 * mm, 100 * mm, 50 * mm]
phantom.translation = [0 * mm, 0 * mm, -500 * mm]
phantom.rotation = m
phantom.material = "G4_AIR"
phantom.color = [1, 0, 1, 1]

# default source for tests (from test42)
source = sim.add_source("PencilBeamSource", "mysource1")
source.mother = "waterbox1"
source.energy.mono = 60 * MeV
source.particle = "proton"
source.position.type = "disc"  # pos = Beam, shape = circle + sigma
source.position.translation = [0 * mm, 0 * mm, -52.5 * cm]
# rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.n = 20000
source.weight = 1
source.direction.partPhSp_x = [
    2.3335754 * mm,
    2.3335754 * mrad,
    0.00078728 * mm * mrad,
    0,
]
source.direction.partPhSp_y = [
    1.96433431 * mm,
    0.00079118 * mrad,
    0.00249161 * mm * mrad,
    0,
]

# add dose actor
dose = sim.add_actor("DoseActor", "doseInYZ_1")
filename = "phantom_a_1.mhd"
dose.output = output_path / filename
dose.mother = "phantom_a_1"
dose.size = [250, 250, 1]
dose.spacing = [0.4, 0.4, 2]
dose.hit_type = "random"


## SECOND SOURCE + DETECTOR ##
# waterbox
# translation and rotation like in the Gate macro
waterbox2 = sim.add_volume("Box", "waterbox2")
waterbox2.size = [10 * cm, 10 * cm, 105 * cm]
waterbox2.translation = [30 * cm, 0 * cm, 52.5 * cm]
waterbox2.material = "Vacuum"
waterbox2.color = [0, 0, 1, 1]

# phantoms
m = Rotation.identity().as_matrix()

phantom2 = sim.add_volume("Box", "phantom_a_2")
phantom2.mother = "waterbox2"
phantom2.size = [100 * mm, 100 * mm, 50 * mm]
phantom2.translation = [0 * mm, 0 * mm, -500 * mm]
phantom2.rotation = m
phantom2.material = "G4_AIR"
phantom2.color = [1, 0, 1, 1]

# default source for tests (from test42)
source2 = sim.add_source("PencilBeamSource", "mysource2")
source2.mother = "waterbox2"
source2.energy.mono = 60 * MeV
source2.particle = "proton"
source2.position.type = "disc"  # pos = Beam, shape = circle + sigma
source2.position.translation = [0 * mm, 0 * mm, -52.5 * cm]
# rotate the disc, equiv to : rot1 0 1 0 and rot2 0 0 1
source2.direction.type = "momentum"
source2.direction.momentum = [0, 0, 1]
source2.n = 20000
source2.weight = 2
source2.direction.partPhSp_x = [
    2.3335754 * mm,
    2.3335754 * mrad,
    0.00078728 * mm * mrad,
    0,
]
source2.direction.partPhSp_y = [
    1.96433431 * mm,
    0.00079118 * mrad,
    0.00249161 * mm * mrad,
    0,
]

# add dose actor
dose2 = sim.add_actor("DoseActor", "doseInYZ_2")
filename = "phantom_a_2.mhd"
dose2.output = output_path / filename
dose2.mother = "phantom_a_2"
dose2.size = [250, 250, 1]
dose2.spacing = [0.4, 0.4, 2]
dose2.hit_type = "random"

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "FTFP_INCLXX_EMZ"
sim.set_cut("world", "all", 1000 * km)

print(sim.dump_sources())

# create output dir, if it doesn't exist
if not os.path.isdir(output_path):
    os.mkdir(output_path)

# start simulation
output = sim.start()

# print results at the end
stat = output.get_actor("Stats")
print(stat)


# ----------------------------------------------------------------------------------------------------------------
# tests

# energy deposition: we expect the edep from source two
# to be double the one of source one

print("\nDifference for EDEP")
mhd_1 = "phantom_a_1.mhd"
mhd_2 = "phantom_a_2.mhd"
test = True
# test = gate.assert_images(
#     output_path / mhd_1,
#     output_path / mhd_2,
#     stat,
#     axis="x",
#     tolerance=50,
#     ignore_value=0,
# )
fig1 = gate.create_2D_Edep_colorMap(output_path / mhd_1, show=False)
fig2 = gate.create_2D_Edep_colorMap(output_path / mhd_2, show=False)

# Total Edep
is_ok = (
    gate.test_weights(
        source2.weight / source.weight,
        output_path / mhd_1,
        output_path / mhd_2,
        thresh=0.2,
    )
    and test
)


gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test037_pet_hits_singles_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test037_pet")

"""
This test considers a PET system (Vereos Philips), with NEMA NECR linear fantom and source (F18).
The digitizer is simplified to:
    1) hits collection
    2) singles are obtained with one simple adder (EnergyWeightedCentroidPosition)

Note that this is not a correct digitizer (no blurring, no noise, no dead-time, etc).

Hits are recorded into the crystal volumes (repeated 23,040 times).
Singles are created by grouping hits from the same event, in the same crystal.

The output is a root file composed of two trees 'Hits' and 'Singles'.
Both are compared to an equivalent legacy Gate simulation.

Salvadori J, Labour J, Odille F, Marie PY, Badel JN, Imbert L, Sarrut D.
Monte Carlo simulation of digital photon counting PET.
EJNMMI Phys. 2020 Apr 25;7(1):23.
doi: 10.1186/s40658-020-00288-w

"""

# create the simulation
sim = gate.Simulation()
crystal = create_pet_simulation(sim, paths)
add_digitizer(sim, paths, "1", crystal)

# timing
sec = gate.g4_units("second")
sim.run_timing_intervals = [[0, 0.00005 * sec]]

# start simulation
output = sim.start()

# print results
stats = output.get_actor("Stats")
print(stats)

# ----------------------------------------------------------------------------------------------------------

# check stats
print()
gate.warning(f"Check stats")
p = paths.gate / "output"
stats_ref = gate.read_stat_file(p / "stats1.txt")
is_ok = gate.assert_stats(stats, stats_ref, 0.025)

# check root hits
hc = output.get_actor("Hits").user_info
f = p / "output1.root"
is_ok = check_root_hits(paths, 1, f, hc.output) and is_ok

# check root singles
sc = output.get_actor("Singles").user_info
is_ok = check_root_singles(paths, 1, f, sc.output) and is_ok

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate


def define_simulation(sim, threads=1):
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")

    ui = sim.user_info
    ui.running_verbose_level = gate.RUN
    ui.g4_verbose = False
    ui.g4_verbose_level = 1
    ui.visu = False
    ui.random_engine = "MersenneTwister"
    ui.random_seed = 123654789
    ui.number_of_threads = threads
    print(ui)

    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    waterbox = sim.add_volume("Box", "Waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    p = sim.get_physics_user_info()
    p.physics_list_name = "QGSP_BERT_EMV"
    cuts = p.production_cuts
    um = gate.g4_units("um")
    cuts.world.gamma = 700 * um
    cuts.world.electron = 700 * um
    cuts.world.positron = 700 * um
    cuts.world.proton = 700 * um

    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 200000 / ui.number_of_threads

    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True


def test_output(output):
    # get output
    paths = gate.get_default_test_paths(__file__, "gate_test004_simulation_stats_actor")

    stats = output.get_actor("Stats")
    print(stats)
    ui = output.simulation.user_info
    stats_ref = gate.read_stat_file(paths.gate_output / "stat.txt")
    stats_ref.counts.run_count = ui.number_of_threads
    is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.01)

    return is_ok



import opengate as gate


class ActorBase(gate.UserElement):
    """
    Store user information about an actor
    """

    @staticmethod
    def set_default_user_info(user_info):
        gate.UserElement.set_default_user_info(user_info)
        # user properties shared for all actors
        user_info.mother = gate.__world_name__
        user_info.filters = []
        user_info.priority = 100

    def __init__(self, user_info):
        # type_name MUST be defined in class that inherit from ActorBase
        super().__init__(user_info)
        # list of filters for this actor
        self.filters_list = []
        # store the output
        self.actor_output = None
        # engines
        self.simulation_engine_wr = None
        self.volume_engine = None
        # sim
        self.simulation = None

    def __del__(self):
        # print("del ActorBase")
        pass

    def __getstate__(self):
        """
        This is important : to get actor's outputs from a simulation run in a separate process,
        the class must be serializable (pickle).
        The engines (volume, actor, etc.) and G4 objects are also removed if exists.
        """
        # do not pickle engines and g4 objects
        for v in self.__dict__:
            if "_engine" in v or "g4_" in v:
                self.__dict__[v] = None
        return self.__dict__

    def initialize(self, simulation_engine_wr=None):
        self.simulation_engine_wr = simulation_engine_wr
        self.volume_engine = self.simulation_engine_wr().volume_engine
        # 'l' must be self to avoid being deleted
        self.filters_list = []
        for f in self.user_info.filters:
            e = gate.new_element(f, self.simulation)
            e.Initialize(f.__dict__)
            self.filters_list.append(e)
        # this is a copy to cpp ('append' cannot be used because fFilters is a std::vector)
        self.fFilters = self.filters_list

    def __str__(self):
        s = f"str ActorBase {self.user_info.name} of type {self.user_info.type_name}"
        return s



import opengate_core as g4
import opengate as gate


class TrackCreatorProcessFilter(g4.GateTrackCreatorProcessFilter, gate.UserElement):
    type_name = "TrackCreatorProcessFilter"

    def set_default_user_info(user_info):
        gate.UserElement.set_default_user_info(user_info)
        # required user info, default values
        user_info.process_name = "none"
        user_info.policy = "keep"  # or "discard"

    def __init__(self, user_info):
        g4.GateTrackCreatorProcessFilter.__init__(self)  # no argument in cpp side
        gate.UserElement.__init__(self, user_info)
        # type_name MUST be defined in class that inherit from a Filter
        if user_info.policy != "keep" and user_info.policy != "discard":
            gate.fatal(
                f'TrackCreatorProcessFilter "{user_info.name}" policy must be either "keep" '
                f'or "discard", while it is "{user_info.policy}"'
            )

    def __getstate__(self):
        # needed to not pickle the g4.GateTrackCreatorProcessFilter
        return self.__dict__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import pathlib

pathFile = pathlib.Path(__file__).parent.resolve()

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_engine = "MersenneTwister"

cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
MeV = gate.g4_units("MeV")

# add a simple volume
waterbox = sim.add_volume("Box", "Waterbox")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
waterbox.material = "G4_WATER"

# default source for tests
# FiXME warning ref not OK (cppSource is not exactly the same)
source = sim.add_source("GenericSource", "Default")
source.particle = "proton"
source.energy.mono = 150 * MeV
source.position.radius = 10 * mm
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.n = 20000

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")

# verbose (WARNING : ui.g4_verbose must be True !)
sim.apply_g4_command("/tracking/verbose 0")
# sim.apply_g4_command("/run/verbose 2")
# sim.apply_g4_command("/event/verbose 2")
# sim.apply_g4_command("/tracking/verbose 1")

print(sim.dump_sources())

# start simulation
output = sim.start()

# get results
stats = output.get_actor("Stats")
print("Simulation seed:", output.current_random_seed)
print(stats)

# gate_test5_proton
# Gate mac/main.mac
print("-" * 80)
stats_ref = gate.read_stat_file(
    pathFile / ".." / "data" / "gate" / "gate_test005_proton" / "output" / "stat.txt"
)
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.15)

gate.test_ok(is_ok)



from .MaterialBuilder import *
from .ElementBuilder import *


class MaterialDatabase:
    """
    Manage a unique list of Geant4 materials and elements.
    The materials/elements are read in a DB txt file or in NIST.
    They are only build on demand, during the geometry Construct
    """

    def __init__(self):
        # list of all db filenames where to read material
        self.filenames = []
        # list of all read material (not build)
        self.material_builders = {}
        self.material_builders_by_filename = {}
        # list of all read element (not build)
        self.element_builders = {}
        self.element_builders_by_filename = {}
        # built materials
        self.g4_materials = {}
        # built elements
        self.g4_elements = {}
        # internal state when reading
        self.current_section = None
        self.current_filename = None
        # specific to NIST materials
        self.g4_NistManager = None
        self.nist_material_names = None
        self.nist_element_names = None

    def __del__(self):
        pass

    def read_from_file(self, filename):
        self.filenames.append(filename)
        self.current_filename = filename
        self.element_builders_by_filename[self.current_filename] = {}
        self.material_builders_by_filename[self.current_filename] = {}
        f = open(filename, "r")
        line = f.readline()
        while line:
            line = line.strip().replace("\t", " ")
            self.read_one_item(f, line)
            line = f.readline()

    def read_one_item(self, f, line):
        # skip empty lines
        if len(line) < 1:
            return
        # skip comment line
        if line[0] == "#":
            return
        # check if the current section change
        w = line.split()[0]
        if w == "[Elements]":
            self.current_section = "element"
            return
        if w == "[Materials]":
            self.current_section = "material"
            return
        if not self.current_section:
            gate.fatal(
                f"Error while reading the file {self.current_filename}, "
                f"current section is {self.current_section}. "
                f"File must start with [Elements] or [Materials]"
            )
        if self.current_section == "element":
            b = ElementBuilder(self)
            b.read(line)
            self.element_builders[b.name] = b
            self.element_builders_by_filename[self.current_filename][b.name] = b
        if self.current_section == "material":
            b = MaterialBuilder(self)
            b.read(f, line)
            self.material_builders[b.name] = b
            self.material_builders_by_filename[self.current_filename][b.name] = b

    def init_NIST(self):
        if self.g4_NistManager is None:
            self.g4_NistManager = g4.G4NistManager.Instance()
            self.nist_material_names = self.g4_NistManager.GetNistMaterialNames()
            self.nist_element_names = self.g4_NistManager.GetNistElementNames()
            self.material_builders_by_filename["NIST"] = self.nist_material_names
            self.element_builders_by_filename["NIST"] = self.nist_element_names

    def FindOrBuildMaterial(self, material_name):
        self.init_NIST()
        # return if already exist
        if material_name in self.g4_materials:
            return self.g4_materials[material_name]
        # we build and store the G4 material if not
        if material_name in self.nist_material_names:
            bm = self.g4_NistManager.FindOrBuildMaterial(material_name)
            self.g4_materials[material_name] = bm
            return bm
        if material_name not in self.material_builders:
            gate.fatal(f'Cannot find nor build material named "{material_name}"')
        bm = self.material_builders[material_name].build()
        self.g4_materials[material_name] = bm
        return bm

    def FindOrBuildElement(self, element_name):
        self.init_NIST()
        # return if already exist
        if element_name in self.g4_elements.keys():
            return self.g4_elements[element_name]
        # we build and store the G4 element if not
        if element_name in self.nist_element_names:
            be = self.g4_NistManager.FindOrBuildElement(element_name)
            self.g4_elements[element_name] = be
            return be
        if element_name not in self.element_builders:
            gate.fatal(f'Cannot find nor build element named "{element_name}"')
        be = self.element_builders[element_name].build()
        self.g4_elements[element_name] = be
        return be

    def get_database_material_names(self, db=None):
        if db is None:
            names = [m for m in self.material_builders]
            return names
        if db not in self.material_builders_by_filename:
            gate.fatal(
                f"The database '{db}' is not in the list of read database: {self.filenames}"
            )
        list = self.material_builders_by_filename[db]
        return [name for name in list]



import opengate as gate
import opengate_core as g4


class DigitizerAdderActor(g4.GateDigitizerAdderActor, gate.ActorBase):
    """
    Equivalent to Gate "adder": gather all hits of an event in the same volume.
    Input: a HitsCollection, need aat least TotalEnergyDeposit and PostPosition attributes
    Output: a Single collections

    Policies:
    - EnergyWinnerPosition: consider position and energy of the hit with the max energy
       for all other attributes (Time, etc.): the value of the winner is used.
    - EnergyWeightedCentroidPosition: computed the energy-weighted centroid position
       for all other attributes (Time, etc.): the value the last seen hit is used.

    """

    type_name = "DigitizerAdderActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        user_info.attributes = []
        user_info.output = "singles.root"
        user_info.input_digi_collection = "Hits"
        user_info.policy = "EnergyWinnerPosition"  # EnergyWeightedCentroidPosition
        user_info.time_difference = False
        user_info.number_of_hits = False
        user_info.skip_attributes = []
        user_info.clear_every = 1e5
        user_info.group_volume = None

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GateDigitizerAdderActor.__init__(self, user_info.__dict__)
        actions = {"StartSimulationAction", "EndSimulationAction"}
        self.AddActions(actions)
        if (
            user_info.policy != "EnergyWinnerPosition"
            and user_info.policy != "EnergyWeightedCentroidPosition"
        ):
            gate.fatal(
                f"Error, the policy for the Adder '{user_info.name}' must be EnergyWinnerPosition or "
                f"EnergyWeightedCentroidPosition, while is is '{user_info.policy}'"
            )

    def __del__(self):
        pass

    def __str__(self):
        s = f"DigitizerAdderActor {self.user_info.name}"
        return s

    def set_group_by_depth(self):
        depth = -1
        if self.user_info.group_volume is not None:
            depth = self.simulation.volume_manager.get_volume_depth(
                self.user_info.group_volume
            )
        self.SetGroupVolumeDepth(depth)

    def StartSimulationAction(self):
        self.set_group_by_depth()
        g4.GateDigitizerAdderActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerAdderActor.EndSimulationAction(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "gate_test009_voxels")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False

# add a material database
sim.add_material_database(paths.data / "GateMaterials.db")

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
mm = gate.g4_units("mm")

#  change world size
world = sim.world
world.size = [1 * m, 1 * m, 1 * m]

# add a simple fake volume to test hierarchy
# translation and rotation like in the Gate macro
fake = sim.add_volume("Box", "fake")
fake.size = [40 * cm, 40 * cm, 40 * cm]
fake.material = "G4_AIR"
fake.color = [1, 0, 1, 1]
fake.rotation = Rotation.from_euler("x", 20, degrees=True).as_matrix()

# image
patient = sim.add_volume("Image", "patient")
patient.image = paths.data / "patient-4mm.mhd"
patient.mother = "fake"
patient.material = "G4_AIR"  # material used by default
patient.voxel_materials = [
    [-2000, -900, "G4_AIR"],
    [-900, -100, "Lung"],
    [-100, 0, "G4_ADIPOSE_TISSUE_ICRP"],
    [0, 300, "G4_TISSUE_SOFT_ICRP"],
    [300, 800, "G4_B-100_BONE"],
    [800, 6000, "G4_BONE_COMPACT_ICRU"],
]
# or alternatively, from a file (like in Gate)
vm = gate.read_voxel_materials(paths.gate_data / "patient-HU2mat-v1.txt")
vm[0][0] = -2000
assert vm == patient.voxel_materials
patient.voxel_materials = vm
# write the image of labels (None by default)
patient.dump_label_image = paths.output / "test009_label.mhd"

# default source for tests
source = sim.add_source("GenericSource", "mysource")
source.energy.mono = 130 * MeV
source.particle = "proton"
source.position.type = "sphere"
source.position.radius = 10 * mm
source.position.translation = [0, 0, -14 * cm]
source.activity = 10000 * Bq
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]

# cuts
c = sim.get_physics_user_info().production_cuts
c.patient.electron = 3 * mm

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test009-edep.mhd"
dose.mother = "patient"
dose.size = [99, 99, 99]
dose.spacing = [2 * mm, 2 * mm, 2 * mm]
dose.img_coord_system = True
dose.translation = [2 * mm, 3 * mm, -2 * mm]
dose.hit_type = "random"

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

# print info
print(sim.dump_volumes())

# verbose
sim.apply_g4_command("/tracking/verbose 0")

# start simulation
output = sim.start(True)

# print results at the end
stat = output.get_actor("Stats")
print(stat)
d = output.get_actor("dose")
print(d)

# tests
stats_ref = gate.read_stat_file(paths.gate_output / "stat.txt")
is_ok = gate.assert_stats(stat, stats_ref, 0.15)
is_ok = is_ok and gate.assert_images(
    paths.gate_output / "output-Edep.mhd",
    paths.output / "test009-edep.mhd",
    stat,
    tolerance=35,
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import os

paths = gate.get_default_test_paths(__file__, "gate_test004_simulation_stats_actor")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = True
ui.visu_type = "vrml"
ui.visu_filename = "geant4VisuFile.wrl"
ui.visu_verbose = True
ui.number_of_threads = 1
ui.random_engine = "MersenneTwister"
ui.random_seed = "auto"

# set the world size like in the Gate macro
m = gate.g4_units("m")
world = sim.world
world.size = [3 * m, 3 * m, 3 * m]

# add a simple waterbox volume
waterbox = sim.add_volume("Box", "Waterbox")
cm = gate.g4_units("cm")
waterbox.size = [40 * cm, 40 * cm, 40 * cm]
waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
waterbox.material = "G4_WATER"

# default source for tests
keV = gate.g4_units("keV")
mm = gate.g4_units("mm")
Bq = gate.g4_units("Bq")
source = sim.add_source("GenericSource", "Default")
source.particle = "gamma"
source.energy.mono = 80 * keV
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
# source.activity = 200000 * Bq
source.activity = 200 * Bq

# runs
sec = gate.g4_units("second")
sim.run_timing_intervals = [[0, 0.5 * sec], [0.5 * sec, 1.0 * sec]]

# add stat actor
sim.add_actor("SimulationStatisticsActor", "Stats")

# start simulation
# sim.apply_g4_command("/run/verbose 1")
output = sim.start()

stats = output.get_actor("Stats")
stats.counts.run_count = 1

# gate_test4_simulation_stats_actor
# Gate mac/main.mac
stats_ref = gate.read_stat_file(paths.gate_output / "stat.txt")
# is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.03)

# gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test014_engine_helpers import *

sim = gate.Simulation()
define_simulation(sim, 3)

# go with a new process that will use 3 threads
se = gate.SimulationEngine(sim, start_new_process=True)
output = se.start()

# get output
is_ok = test_output(output)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test014_engine_helpers import *

sim = gate.Simulation()
define_simulation(sim, 5)

# go without sub process, but with multithread
se = gate.SimulationEngine(sim, start_new_process=False)
output = se.start()

# get output
is_ok = test_output(output)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4


def create_pl_sim():
    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = True
    ui.g4_verbose_level = 1
    ui.visu = False
    ui.random_engine = "MersenneTwister"
    ui.random_seed = 123456987

    # set the world size like in the Gate macro
    m = gate.g4_units("m")
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]

    # add a simple waterbox volume
    waterbox = sim.add_volume("Box", "waterbox")
    cm = gate.g4_units("cm")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    # add a daughter (in wb)
    b1 = sim.add_volume("Box", "b1")
    b1.mother = "waterbox"
    b1.size = [4 * cm, 4 * cm, 4 * cm]
    b1.translation = [5 * cm, 5 * cm, 0 * cm]
    b1.material = "G4_Pd"

    # add another box (in world)
    b2 = sim.add_volume("Box", "b2")
    b2.size = [4 * cm, 4 * cm, 4 * cm]
    b2.translation = [0 * cm, 0 * cm, 0 * cm]
    b2.material = "G4_LUNG_ICRP"

    # physics
    mm = gate.g4_units("mm")
    eV = gate.g4_units("eV")
    MeV = gate.g4_units("MeV")
    p = sim.get_physics_user_info()
    p.energy_range_min = 250 * eV
    p.energy_range_max = 15 * MeV

    # print info about physics
    print("Phys list:", p)
    # print("Phys list param:")
    # print(p.g4_em_parameters.ToString()) # no more available before init
    print("Available phys lists:")
    print(sim.physics_manager.dump_available_physics_lists())

    # default source for tests
    MeV = gate.g4_units("MeV")
    Bq = gate.g4_units("Bq")

    source = sim.add_source("GenericSource", "gamma")
    source.particle = "gamma"
    source.energy.mono = 10 * MeV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 10000 * Bq

    source = sim.add_source("GenericSource", "ion1")
    source.particle = "ion 9 18"  # or F18 or Fluorine18
    source.position.type = "sphere"
    source.position.translation = [10 * mm, 10 * mm, 20 * mm]
    source.position.radius = 3 * mm
    source.direction.type = "iso"
    source.activity = 2000 * Bq

    source = sim.add_source("GenericSource", "ion2")
    source.particle = "ion 53 124"  # 53 124 0 0       # Iodine 124
    source.position.type = "sphere"
    source.position.translation = [-10 * mm, -10 * mm, -40 * mm]
    source.position.radius = 1 * mm
    source.direction.type = "iso"
    source.activity = 2000 * Bq

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    return sim


def phys_em_parameters(p):
    # must be done after the initialization
    em = g4.G4EmParameters.Instance()  # p.g4_em_parameters
    em.SetFluo(True)
    em.SetAuger(True)
    em.SetAugerCascade(True)
    em.SetPixe(True)
    em.SetDeexcitationIgnoreCut(True)

    # is this needed to do like Gate ? (unsure)
    em.SetDeexActiveRegion("world", True, True, True)
    # em.SetDeexActiveRegion("waterbox", True, True, True)
    # em.SetDeexActiveRegion("b1", True, True, True)
    # em.SetDeexActiveRegion("b2", True, True, True)
